"""
Instrument driver for the Rohde&Schwarz SMBV100A RF Signal Generator
"""

import logging
from typing import Union
import numpy as np

from qmi.core.rpc import rpc_method
from qmi.instruments.rohde_schwarz.rs_base_signal_gen import RohdeSchwarz_Base

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


def _create_list(start: Union[float, int], stop: Union[float, int], unit: str, number_of_steps: int) -> str:
    """
    Create a list of values given the start, stop and step values.
    Append the unit to the values in the list.

    Parameters:
        start:             The starting value of the list.
        stop:              The ending value of the list.
        unit:              The unit of the values.
        number_of_steps:   The number of steps in the list.
    """
    array: np.ndarray = np.linspace(start, stop, number_of_steps)
    return ",".join(map(lambda l: str(l) + unit, list(array)))


class RohdeSchwarz_SMBV100A(RohdeSchwarz_Base):
    """Instrument driver for the Rohde&Schwarz SMBV100A RF Signal Generator."""

    def _set_sig_for_iq_mod(self, sig: str) -> None:
        """
        Sets the input signal for I/Q modulation.
        This private version of the method does not perform any
        calibration or error checks, since it is used by the
        calibration method.

        Parameters:
            sig: "bas", "anal" or "diff"
        """
        options = ["BAS", "ANAL", "DIFF"]
        sig = self._is_valid_param(sig, options)
        _logger.info(__name__ + " : setting input signal for IQ modulation to '%s'" % sig)
        self._scpi_protocol.write(f":IQ:SOUR {sig}")

    @rpc_method
    def set_sig_for_iq_mod(self, sig: str) -> None:
        """
        Sets the input signal for I/Q modulation.

        Parameters:
            sig: "bas", "anal" or "diff"
        """
        self._check_calibrating()
        self._set_sig_for_iq_mod(sig)
        self._check_error()

    @rpc_method
    def get_external_reference_frequency(self) -> str:
        """Return the currently configured external reference input frequency.

        Possible values: "5MHZ", "10MHZ".
        """
        return self._get_external_reference_frequency()

    @rpc_method
    def set_external_reference_frequency(self, frequency: str) -> None:
        """Configure the external reference input frequency.

        Parameters:
            frequency: desired frequency (accepted values: "5MHZ", "10MHZ"); see also
                       get_external_reference_frequency().
        """
        freq_options = ["5MHZ", "10MHZ"]
        self._set_external_reference_frequency(frequency, freq_options)

    def _set_iq(self, toggle: bool) -> None:
        """
        Activates external I/Q modulation with analog signal.
        This private version of the method does not perform any
        calibration or error checks, since it is used by the
        calibration method.

        Parameters:
            toggle: boolean flag to turn IQ modulation on/off
        """
        iq = "ON" if toggle else "OFF"
        _logger.info(__name__ + " : setting external IQ modulation to '%s'" % iq)
        self._set_sig_for_iq_mod("ANAL")
        self._scpi_protocol.write(f":IQ:STAT {1 if toggle else 0}")

    @rpc_method
    def set_iq(self, toggle: bool) -> None:
        """
        Activates external I/Q modulation with analog signal.

        Parameters:
            toggle: boolean flag to turn IQ modulation on/off
        """
        self._set_iq(toggle)

    @rpc_method
    def set_freq_mode(self, mode: str) -> None:
        """
        Sets RF frequency mode to list, sweep or fixed.
        CW and FIX are the same

        Parameters:
            mode: "fix", "list", "swe" or "cw"
        """
        options = ["FIX", "LIST", "SWE", "CW"]
        mode = self._is_valid_param(mode, options)
        self._check_calibrating()
        _logger.info(__name__ + " : setting RF frequency mode to '%s'" % mode)
        self._scpi_protocol.write(f"SOUR:FREQ:MODE {mode}")
        self._check_error()

    @rpc_method
    def set_sweep_frequency_start(self, frequency: float) -> None:
        """
        Set start frequency of sweep.

        Parameters:
            frequency: frequency in Hz
        """
        self._check_calibrating()
        _logger.info(__name__ + " : setting sweep frequency start to %s Hz" % frequency)
        self._scpi_protocol.write(f"FREQ:STAR {frequency} Hz")
        self._check_error()

    @rpc_method
    def set_sweep_frequency_stop(self, frequency: float) -> None:
        """
        Set stop frequency of sweep.

        Parameters:
            frequency: frequency in Hz
        """
        self._check_calibrating()
        _logger.info(__name__ + " : setting sweep frequency stop to %s Hz" % frequency)
        self._scpi_protocol.write(f"FREQ:STOP {frequency} Hz")
        self._check_error()

    @rpc_method
    def set_sweep_frequency_step(self, frequency: float) -> None:
        """
        Set step frequency of sweep.

        Parameters:
            frequency: frequency in Hz
        """
        self._check_calibrating()
        _logger.info(__name__ + " : setting sweep frequency step to %s Hz" % frequency)
        self._scpi_protocol.write(f"SWE:STEP:LIN {frequency} Hz")
        self._check_error()

    @rpc_method
    def start_calibration(self, all: bool = False, cal_iqm: bool = True) -> None:
        """
        Parameters:
            all:     Boolean to select if to calibrate everything.
            cal_iqm: Boolean to select if to calibrate IQ modulator. Has effect only if 'all' is False.
        """
        super().start_calibration()
        if all:
            self._scpi_protocol.write("CAL:ALL:MEAS?")

        else:
            self._scpi_protocol.write("CAL:FREQ:MEAS?")
            _logger.info("Frequency calibrated")
            self._scpi_protocol.write("CAL:LEV:MEAS?")
            _logger.info("Level calibrated")
            if cal_iqm:
                self._set_iq(True)
                self._scpi_protocol.write("CAL:IQM:LOC?")
                _logger.info("IQ modulator calibrated")

    @rpc_method
    def enable_list_mode(self) -> None:
        """
        Convenience function to set RF frequency to list mode.
        """
        self.set_freq_mode("LIST")

    @rpc_method
    def set_list_processing_mode(self, mode: str) -> None:
        """
        Selects how the list is to be processed.

        Parameters:
            mode: "auto" or "step"
        """
        options = ["AUTO", "STEP"]
        mode = self._is_valid_param(mode, options)
        self._check_calibrating()
        _logger.info(__name__ + " : setting list processing mode to '%s'" % mode)
        self._scpi_protocol.write(f"LIST:MODE {mode}")
        self._check_error()

    @rpc_method
    def enable_list_step_mode(self):
        """
        Convenience function to select step-by-step processing of the list.
        """
        self.set_list_processing_mode("STEP")

    @rpc_method
    def set_trigger_source_processing_lists(self, mode: str) -> None:
        """
        Sets the trigger source processing lists.

        Parameters:
            mode: "auto", "imm", "sing" or "ext"
        """
        options = ["AUTO", "IMM", "SING", "EXT"]
        mode = self._is_valid_param(mode, options)
        self._check_calibrating()
        _logger.info(__name__ + " : setting trigger source processing list to '%s'" % mode)
        self._scpi_protocol.write(f"LIST:TRIG:SOUR {mode}")
        self._check_error()

    @rpc_method
    def set_list_ext_trigger_source(self):
        """
        Selects triggering by means of the external trigger.
        """
        self.set_trigger_source_processing_lists("EXT")

    @rpc_method
    def set_freq_sweep_mode(self, mode: str):
        """
        Selects the frequency sweep mode.

        Parameters:
            mode: "auto" or "step"
        """
        options = ["STEP", "AUTO"]
        mode = self._is_valid_param(mode, options)
        self._check_calibrating()
        _logger.info(__name__ + " : setting frequency sweep mode to '%s'" % mode)
        self._scpi_protocol.write(f"SOUR:SWE:FREQ:MODE {mode}")
        self._check_error()

    @rpc_method
    def set_freq_sweep_spacing_mode(self, mode: str):
        """
        Selects the frequency sweep spacing mode.

        Parameters:
            mode: "log" or "lin"
        """
        options = ["LIN", "LOG"]
        mode = self._is_valid_param(mode, options)
        self._check_calibrating()
        _logger.info(__name__ + " : setting frequency sweep spacing mode to '%s'" % mode)
        self._scpi_protocol.write(f"SOUR:SWE:FREQ:SPAC {mode}")
        self._check_error()

    @rpc_method
    def set_trig_source_freq_sweep(self, mode: str):
        """
        Sets the trigger source for the RF frequency sweep.

        Parameters:
            mode: "auto", "sing", "ext" or "eaut"
        """
        options = ["AUTO", "SING", "EXT", "EAUT"]
        mode = self._is_valid_param(mode, options)
        self._check_calibrating()
        _logger.info(__name__ + " : setting trigger source for frequency sweep to '%s'" % mode)
        self._scpi_protocol.write(f"TRIG:FSW:SOUR {mode}")
        self._check_error()

    @rpc_method
    def enable_ext_freq_sweep_mode(self):
        """
        Convenience function to generate sweep signal step-by-step, manually triggered.
        To trigger a sweep step, apply an external trigger signal. The step
        width corresponds to the step width set for the rotary knob.
        """
        self.set_freq_sweep_mode("STEP")
        self.set_freq_sweep_spacing_mode("LIN")
        self.set_trig_source_freq_sweep("EXT")
        self.set_freq_mode("SWE")

    @rpc_method
    def reset_sweep(self):
        """
        Resets all active sweeps to the starting points.
        """
        self._check_calibrating()
        _logger.info(__name__ + " : resetting all active sweeps to the starting points.")
        self._scpi_protocol.write("SOUR:SWE:RES:ALL")
        self._check_error()

    @rpc_method
    def reset_list_mode(self):
        """
        Resets the list to the starting point.
        """
        self._check_calibrating()
        _logger.info(__name__ + " : resetting the list to the starting point")
        self._scpi_protocol.write("SOUR:LIST:RES")
        self._check_error()

    @rpc_method
    def learn_list(self):
        """
        Learns the selected list to determine the hardware setting for
        all list entries. The results are saved with the list. When
        the list is activated the first time, these settings are
        calculated automatically.
        """
        self._check_calibrating()
        _logger.info(__name__ + " : learning the list")
        self._scpi_protocol.write("SOUR:LIST:LEAR")
        self._check_error()

    # TODO: find documentation for this
    @rpc_method
    def reset_list(self):
        self._check_calibrating()
        self._scpi_protocol.write("ABOR:LIST")
        self._check_error()

    @rpc_method
    def load_fplist(
            self,
            fstart: float,
            fstop: float,
            funit: str,
            number_of_steps: int,
            pstart: float,
            pstop: float,
            punit: str
    ):
        """
        Loads a frequency and power list.

        Parameters:
            fstart:             The starting frequency of the list.
            fstop:              The ending frequency of the list.
            funit:              The unit of the frequency.
            number_of_steps:    The number of steps in the list.
            pstart:             The starting power of the list.
            pstop:              The ending power of the list.
            punit:              The unit of the power.
        """
        self._check_calibrating()
        # create new list
        self._scpi_protocol.write(f"SOUR:LIST:SEL 'list_{fstart}_{fstop}_{number_of_steps}'")

        # create both frequency and power lists and format as comma separated values
        flist = _create_list(fstart, fstop, funit, number_of_steps)
        plist = _create_list(pstart, pstop, punit, number_of_steps)

        # write the lists to the instrument
        self._scpi_protocol.write("SOUR:LIST:FREQ " + flist)
        self._scpi_protocol.write("SOUR:LIST:POW " + plist)
        self._check_error()



"""
Instrument driver for the Newport SMC100PP motion controller.
"""
import logging
from time import sleep
from typing import Dict, Optional

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.rpc import rpc_method
from qmi.instruments.newport.actuators import LinearActuator
from qmi.instruments.newport.single_axis_motion_controller import Newport_SingleAxisMotionController

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class Newport_SMC100PP(Newport_SingleAxisMotionController):
    """Instrument driver for the Newport SMC100PP servo motion controller."""

    def __init__(self,
                 context: QMI_Context,
                 name: str,
                 transport: str,
                 serial: str,
                 actuators: Dict[Optional[int], LinearActuator],
                 baudrate: int = 57600) -> None:
        """Initialize driver.

        Parameters:
            name:       Name for this instrument instance.
            transport:  QMI transport descriptor to connect to the instrument.
            serial:     The serial number of the instrument.
            actuators:  The linear actuators that this controller will drive. Each controller address
                        drives a linear actuator. The key of the dictionary is the controller address
                        and the value is the actuator that it drives.
            baudrate:   The baudrate of the instrument. Defaults to 57600.
        """
        super().__init__(context, name, transport, serial,
                         actuators, baudrate)

    @rpc_method
    def get_micro_step_per_full_step_factor(self, controller_address: Optional[int] = None) -> float:
        """
        Get the micro-step per full step factor of the motor configuration.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.

        Returns:
            factor: The micro-step per full step factor as float.
        """
        self.controller_address = controller_address
        _logger.info(
            "Getting the micro-step per full step factor of instrument [%s]", self._name)
        # instrument must be in CONFIGURATION state to get the full step factor.
        self._enter_configuration_state()
        factor = self._scpi_protocol.ask(
            self._build_command("FRM?"))
        self._check_error()
        self._exit_configuration_state()
        return float(factor[4:])

    @rpc_method
    def set_micro_step_per_full_step_factor(self, factor: int, controller_address: Optional[int] = None) -> None:
        """
        Set the micro-step per full step factor of the motor configuration.

        Parameters:
            factor:             Micro-step per full step factor.
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        if factor <= 0 or factor > 2000:
            raise QMI_InstrumentException(
                f"Provided value {factor} not in valid range 0 < factor <= 2000.")

        self.controller_address = controller_address
        _logger.info(
            "Setting the micro-step per full step factor of instrument [%s] to [%i]", self._name, factor)
        # instrument must be in CONFIGURATION state to set the full step factor.
        self._enter_configuration_state()
        self._scpi_protocol.write(self._build_command(
            "FRM", factor))
        sleep(self.COMMAND_EXEC_TIME)
        self._check_error()
        self._exit_configuration_state()

    @rpc_method
    def get_motion_distance_per_full_step(self, controller_address: Optional[int] = None) -> float:
        """
        Get the motion distance per motor’s full step.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.

        Returns:
            m_dist: Motion distance as float.
        """
        self.controller_address = controller_address
        _logger.info(
            "Getting the motion distance per motor’s full step value of instrument [%s]", self._name)
        # instrument must be in CONFIGURATION state to get the motion distance.
        self._enter_configuration_state()
        m_dist = self._scpi_protocol.ask(
            self._build_command("FRS?"))
        self._check_error()
        self._exit_configuration_state()
        return float(m_dist[4:])

    @rpc_method
    def set_motion_distance_per_full_step(self, m_dist: float, controller_address: Optional[int] = None) -> None:
        """
        Set the motion distance per motor’s full step.

        Parameters:
            m_dist:             Motion distance per motor's full step.
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        if m_dist <= self.MIN_FLOAT_LIMIT or m_dist >= self.MAX_FLOAT_LIMIT:
            raise QMI_InstrumentException(
                f"Provided value {m_dist} not in valid range {self.MIN_FLOAT_LIMIT} < m_dist < {self.MAX_FLOAT_LIMIT}."
            )

        self.controller_address = controller_address
        _logger.info(
            "Setting the motion distance per motor’s full step of instrument [%s] to [%f]", self._name, m_dist)
        # instrument must be in CONFIGURATION state to set the motion distance.
        self._enter_configuration_state()
        self._scpi_protocol.write(self._build_command(
            "FRS", m_dist))
        self._check_error()
        self._exit_configuration_state()

    @rpc_method
    def get_base_velocity(self, controller_address: Optional[int] = None) -> float:
        """
        Get the profile generator base velocity.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.

        Returns:
            base_velocity: Base velocity as float.
        """
        self.controller_address = controller_address
        _logger.info(
            "Getting the profile generator base velocity of instrument [%s]", self._name)
        # instrument must be in CONFIGURATION, DISABLE or READY state to get the base velocity.
        try:
            config_state = False
            self._state_ready_check("base velocity")

        except QMI_InstrumentException:
            self._enter_configuration_state()
            config_state = True

        base_velocity = self._scpi_protocol.ask(
            self._build_command("VB?"))
        self._check_error()
        if config_state:
            self._exit_configuration_state()

        else:
            self._exit_disable_state()

        return float(base_velocity[3:])

    @rpc_method
    def set_base_velocity(
            self, base_velocity: float, persist: Optional[bool] = False, controller_address: Optional[int] = None
    ) -> None:
        """
        Set the profile generator base velocity. It must not be larger than the maximum velocity set by VA command.

        Parameters:
            base_velocity:      New profile generator base velocity.
            persist:            Flag to indicate if the velocity should be persisted to the controller's memory, so it
                                is still available after powering down the controller. When not persisted, the maximum
                                base velocity that can be set is the one stored in the controller's memory.
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        self.controller_address = controller_address
        if persist:
            self._enter_configuration_state()

        else:
            self._state_ready_check("base velocity")

        response = self._scpi_protocol.ask(
            self._build_command("VA?"))
        sleep(self.COMMAND_EXEC_TIME)
        self._check_error()
        velocity = float(response[3:])
        if base_velocity < 0 or base_velocity > velocity:
            if persist:
                self._exit_configuration_state()
            raise QMI_InstrumentException(
                f"Provided value {base_velocity} not in valid range 0 <= base_velocity <= {velocity}.")

        _logger.info(
            "Setting the profile generator base velocity of instrument [%s] to [%f]", self._name, base_velocity)
        self._scpi_protocol.write(self._build_command(
            "VB", base_velocity))
        sleep(self.COMMAND_EXEC_TIME)
        self._check_error()
        if persist:
            self._exit_configuration_state()



"""Unit-tests for Rohde&Schwarz SMBV100a."""
from typing import cast
import unittest
from unittest.mock import Mock, call, patch

import qmi
from qmi.core.exceptions import QMI_InstrumentException, QMI_TimeoutException
from qmi.core.transport import QMI_TcpTransport
from qmi.instruments.rohde_schwarz import RohdeSchwarz_Smbv100a


class TestSMBV100A(unittest.TestCase):

    def setUp(self):
        qmi.start("TestSMBV100AContext")
        # Add patches
        patcher = patch('qmi.instruments.rohde_schwarz.rs_base_signal_gen.create_transport', spec=QMI_TcpTransport)
        self._transport_mock: Mock = patcher.start().return_value
        self.addCleanup(patcher.stop)
        patcher = patch('qmi.instruments.rohde_schwarz.rs_base_signal_gen.ScpiProtocol', autospec=True)
        self._scpi_mock: Mock = patcher.start().return_value
        self.addCleanup(patcher.stop)
        # Make DUT
        self.instr: RohdeSchwarz_Smbv100a = qmi.make_instrument("SMBV100a", RohdeSchwarz_Smbv100a, "")
        self.instr = cast(RohdeSchwarz_Smbv100a, self.instr)
        self.instr.open()

    def tearDown(self):
        self.instr.close()
        qmi.stop()

    def test_start_calibration_for_all(self):
        """Test start calibration for all."""
        self.instr.start_calibration(all=True)

        self._scpi_mock.write.assert_called_once_with("CAL:ALL:MEAS?")

    def test_start_calibration_for_frequency_and_level(self):
        """Test start calibration for frequency and level."""
        self.instr.start_calibration(cal_iqm=False)
        expected_calls = [
            call("CAL:FREQ:MEAS?"),
            call("CAL:LEV:MEAS?")
        ]

        self._scpi_mock.write.assert_has_calls(expected_calls)
    
    def test_start_calibration_for_frequency_level_and_iqm(self):
        """Test start calibration for frequency, level and IQM."""
        self._scpi_mock.ask.return_value = "0,\"No error\""
        self.instr.start_calibration()
        expected_calls = [
            call("CAL:FREQ:MEAS?"),
            call("CAL:LEV:MEAS?"),
            call(":IQ:SOUR ANAL"),
            call(":IQ:STAT 1"),
            call("CAL:IQM:LOC?")
        ]

        self._scpi_mock.write.assert_has_calls(expected_calls)

    def test_check_if_calibrating(self):
        """Test that ongoing calibration inhibits interactions with the instrument."""
        # Device doesn't respond during calibration
        self._transport_mock.read_until.side_effect = QMI_TimeoutException
        self._scpi_mock.ask.return_value = "0,\"No error\""

        self.instr.start_calibration()

        with self.assertRaises(QMI_InstrumentException):
            self.instr.get_idn()

    def test_get_set_external_ref_frequency(self):
        """Test get/set reference frequency."""
        # Test get.
        value = "10MHZ"
        self._scpi_mock.ask.return_value = value

        result = self.instr.get_external_reference_frequency()

        self._scpi_mock.ask.assert_called_once_with(":ROSC:EXT:FREQ?")
        self.assertEqual(result, value)

        # Test set.
        self._scpi_mock.ask.return_value = "0,\"No error\""
        for target_value in ("5MHZ", "10MHZ", "5MHz", "10mhz"):
            self.instr.set_external_reference_frequency(target_value)

            self._scpi_mock.write.assert_called_once_with(f":ROSC:EXT:FREQ {target_value.upper()}")
            self._scpi_mock.write.reset_mock()

        # Test invalid value.
        with self.assertRaises(ValueError):
            self.instr.set_external_reference_frequency("10000MHZ")

    def test_set_sig_for_iq_mod_with_valid_signal(self):
        """Test setting the input signal for I/Q modulation with valid inputs."""
        valid_sigs = ['BAS', 'ANAL', 'DIFF']
        for sig in valid_sigs:
            self._scpi_mock.ask.return_value = "0,\"No error\""
            with self.subTest("Testing %s" % sig, sig=sig):
                self.instr.set_sig_for_iq_mod(sig)
                self._scpi_mock.write.assert_called_once_with(f":IQ:SOUR {sig}")
            self._scpi_mock.reset_mock()
    
    def test_set_sig_for_iq_mod_with_invalid_signal(self):
        """Test setting the input signal for I/Q modulation with invalid inputs."""
        valid_sigs = ['SWE', 'CW', 'FIX']
        for sig in valid_sigs:
            self._scpi_mock.ask.return_value = "Values that can be set are BAS,ANAL,DIFF"
            with self.subTest("Testing %s" % sig, sig=sig):
                with self.assertRaises(ValueError):
                    self.instr.set_sig_for_iq_mod(sig)
            self._scpi_mock.reset_mock()
    
    def test_set_iq(self):
        """Test activating/deactivating I/Q modulation with analog signal."""
        self._scpi_mock.ask.return_value = "0,\"No error\""
        # test activating
        expected_calls = [
            call(":IQ:SOUR ANAL"),
            call(":IQ:STAT 1")
        ]

        self.instr.set_iq(True)
        self._scpi_mock.write.assert_has_calls(expected_calls)
        self._scpi_mock.reset_mock()
        # test deactivating
        expected_calls = [
            call(":IQ:SOUR ANAL"),
            call(":IQ:STAT 0")
        ]

        self.instr.set_iq(False)
        self._scpi_mock.write.assert_has_calls(expected_calls)

    def test_set_freq_mode_with_valid_signal(self):
        """Test setting RF frequency mode with valid inputs."""
        valid_modes = ['FIX', 'LIST', 'SWE', 'CW']
        for mode in valid_modes:
            self._scpi_mock.ask.return_value = "0,\"No error\""
            with self.subTest("Testing %s" % mode, mode=mode):
                self.instr.set_freq_mode(mode)
                self._scpi_mock.write.assert_called_once_with(f"SOUR:FREQ:MODE {mode}")
            self._scpi_mock.reset_mock()
    
    def test_set_freq_mode_with_invalid_signal(self):
        """Test setting RF frequency mode with invalid inputs."""
        valid_modes = ['BAS', 'ANAL', 'DIFF']
        for mode in valid_modes:
            self._scpi_mock.ask.return_value = "Values that can be set are FIX,LST,SWE,CW"
            with self.subTest("Testing %s" % mode, mode=mode):
                with self.assertRaises(ValueError):
                    self.instr.set_freq_mode(mode)
            self._scpi_mock.reset_mock()

    def test_set_sweep_frequency_start(self):
        """Test setting start frequency for sweep"""
        self._scpi_mock.ask.return_value = "0,\"No error\""
        freq = 10.5
        self.instr.set_sweep_frequency_start(freq)
        self._scpi_mock.write.assert_called_once_with(f"FREQ:STAR {freq} Hz")

    def test_set_sweep_frequency_stop(self):
        """Test setting stop frequency for sweep"""
        self._scpi_mock.ask.return_value = "0,\"No error\""
        freq = 100.5
        self.instr.set_sweep_frequency_stop(freq)
        self._scpi_mock.write.assert_called_once_with(f"FREQ:STOP {freq} Hz")

    def test_set_sweep_frequency_step(self):
        """Test setting frequency step for sweep"""
        self._scpi_mock.ask.return_value = "0,\"No error\""
        freq = 2
        self.instr.set_sweep_frequency_step(freq)
        self._scpi_mock.write.assert_called_once_with(f"SWE:STEP:LIN {freq} Hz")

    def test_enable_list_mode(self):
        """Test enabling of list mode."""
        self._scpi_mock.ask.return_value = "0,\"No error\""
        self.instr.enable_list_mode()

        self._scpi_mock.write.assert_called_once_with("SOUR:FREQ:MODE LIST")

    def test_set_list_processing_mode_with_valid_signal(self):
        """Test setting RF frequency mode with valid inputs."""
        valid_modes = ['AUTO', 'STEP']
        for mode in valid_modes:
            self._scpi_mock.ask.return_value = "0,\"No error\""
            with self.subTest("Testing %s" % mode, mode=mode):
                self.instr.set_list_processing_mode(mode)
                self._scpi_mock.write.assert_called_once_with(f"LIST:MODE {mode}")
            self._scpi_mock.reset_mock()
    
    def test_set_list_processing_mode_with_invalid_signal(self):
        """Test setting RF frequency mode with invalid inputs."""
        valid_modes = ['BAS', 'ANAL', 'DIFF']
        for mode in valid_modes:
            self._scpi_mock.ask.return_value = "Values that can be set are AUTO,STEP"
            with self.subTest("Testing %s" % mode, mode=mode):
                with self.assertRaises(ValueError):
                    self.instr.set_list_processing_mode(mode)
            self._scpi_mock.reset_mock()

    def test_enable_list_step_mode(self):
        """Test enabling of step-by-step processing of list."""
        self._scpi_mock.ask.return_value = "0,\"No error\""
        self.instr.enable_list_step_mode()

        self._scpi_mock.write.assert_called_once_with("LIST:MODE STEP")

    def test_set_trigger_source_processing_lists_with_valid_signal(self):
        """Test setting the trigger source processing lists with valid inputs."""
        valid_modes = ['AUTO', 'IMM', 'SING', 'EXT']
        for mode in valid_modes:
            self._scpi_mock.ask.return_value = "0,\"No error\""
            with self.subTest("Testing %s" % mode, mode=mode):
                self.instr.set_trigger_source_processing_lists(mode)
                self._scpi_mock.write.assert_called_once_with(f"LIST:TRIG:SOUR {mode}")
            self._scpi_mock.reset_mock()
    
    def test_set_trigger_source_processing_lists_with_invalid_signal(self):
        """Test setting the trigger source processing lists with invalid inputs."""
        valid_modes = ['BAS', 'ANAL', 'DIFF']
        for mode in valid_modes:
            self._scpi_mock.ask.return_value = "Values that can be set are AUTO,STEP"
            with self.subTest("Testing %s" % mode, mode=mode):
                with self.assertRaises(ValueError):
                    self.instr.set_trigger_source_processing_lists(mode)
            self._scpi_mock.reset_mock()

    def test_set_list_ext_trigger_source(self):
        """Test triggering of list externally."""
        self._scpi_mock.ask.return_value = "0,\"No error\""
        self.instr.set_list_ext_trigger_source()

        self._scpi_mock.write.assert_called_once_with("LIST:TRIG:SOUR EXT")

    def test_set_freq_sweep_mode_with_valid_signal(self):
        """Test setting the frequency sweep mode with valid inputs."""
        valid_modes = ['STEP', 'AUTO']
        for mode in valid_modes:
            self._scpi_mock.ask.return_value = "0,\"No error\""
            with self.subTest("Testing %s" % mode, mode=mode):
                self.instr.set_freq_sweep_mode(mode)
                self._scpi_mock.write.assert_called_once_with(f"SOUR:SWE:FREQ:MODE {mode}")
            self._scpi_mock.reset_mock()
    
    def test_set_freq_sweep_mode_with_invalid_signal(self):
        """Test setting the frequency sweep mode with invalid inputs."""
        valid_modes = ['BAS', 'ANAL', 'DIFF']
        for mode in valid_modes:
            self._scpi_mock.ask.return_value = "Values that can be set are AUTO,STEP"
            with self.subTest("Testing %s" % mode, mode=mode):
                with self.assertRaises(ValueError):
                    self.instr.set_freq_sweep_mode(mode)
            self._scpi_mock.reset_mock()

    def test_set_freq_sweep_spacing_mode_with_valid_signal(self):
        """Test setting the frequency sweep spacing mode with valid inputs."""
        valid_modes = ['LIN', 'LOG']
        for mode in valid_modes:
            self._scpi_mock.ask.return_value = "0,\"No error\""
            with self.subTest("Testing %s" % mode, mode=mode):
                self.instr.set_freq_sweep_spacing_mode(mode)
                self._scpi_mock.write.assert_called_once_with(f"SOUR:SWE:FREQ:SPAC {mode}")
            self._scpi_mock.reset_mock()
    
    def test_set_freq_sweep_spacing_mode_with_invalid_signal(self):
        """Test setting the frequency sweep spacing mode with invalid inputs."""
        valid_modes = ['BAS', 'ANAL', 'DIFF']
        for mode in valid_modes:
            self._scpi_mock.ask.return_value = "Values that can be set are LIN,LOG"
            with self.subTest("Testing %s" % mode, mode=mode):
                with self.assertRaises(ValueError):
                    self.instr.set_freq_sweep_spacing_mode(mode)
            self._scpi_mock.reset_mock()

    def test_set_trig_source_freq_sweep_with_valid_signal(self):
        """Test setting the trigger source for the RF frequency sweep with valid inputs."""
        valid_modes = ['AUTO', 'SING', 'EXT', 'EAUT']
        for mode in valid_modes:
            self._scpi_mock.ask.return_value = "0,\"No error\""
            with self.subTest("Testing %s" % mode, mode=mode):
                self.instr.set_trig_source_freq_sweep(mode)
                self._scpi_mock.write.assert_called_once_with(f"TRIG:FSW:SOUR {mode}")
            self._scpi_mock.reset_mock()
    
    def test_set_trig_source_freq_sweep_with_invalid_signal(self):
        """Test setting the trigger source for the RF frequency sweep with invalid inputs."""
        valid_modes = ['BAS', 'ANAL', 'DIFF']
        for mode in valid_modes:
            self._scpi_mock.ask.return_value = "Values that can be set are AUTO,SING,EXT,EAUT"
            with self.subTest("Testing %s" % mode, mode=mode):
                with self.assertRaises(ValueError):
                    self.instr.set_trig_source_freq_sweep(mode)
            self._scpi_mock.reset_mock()

    def test_set_freq_mode_sig_gen_with_valid_signal(self):
        """Test setting the trigger source for the RF frequency sweep with valid inputs."""
        valid_modes = ['CW', 'FIX', 'SWE', 'LIST']
        for mode in valid_modes:
            self._scpi_mock.ask.return_value = "0,\"No error\""
            with self.subTest("Testing %s" % mode, mode=mode):
                self.instr.set_freq_mode(mode)
                self._scpi_mock.write.assert_called_once_with('SOUR:FREQ:MODE %s' % mode)
            self._scpi_mock.reset_mock()
    
    def test_set_freq_mode_sig_gen_with_invalid_signal(self):
        """Test setting the trigger source for the RF frequency sweep with invalid inputs."""
        valid_modes = ['BAS', 'ANAL', 'DIFF']
        for mode in valid_modes:
            self._scpi_mock.ask.return_value = "Values that can be set are CW,FIX,SWE,LIST"
            with self.subTest("Testing %s" % mode, mode=mode):
                with self.assertRaises(ValueError):
                    self.instr.set_freq_mode(mode)
            self._scpi_mock.reset_mock()

    def test_enable_ext_freq_sweep_mode(self):
        """Test enabling of sweep mode with externally triggered input."""
        self._scpi_mock.ask.return_value = "0,\"No error\""
        self.instr.enable_ext_freq_sweep_mode()

        expected_calls = [
            call("SOUR:SWE:FREQ:MODE STEP"),
            call("SOUR:SWE:FREQ:SPAC LIN"),
            call("TRIG:FSW:SOUR EXT"),
            call("SOUR:FREQ:MODE SWE")
        ]

        self._scpi_mock.write.assert_has_calls(expected_calls)

    def test_reset_sweep(self):
        """Test resetting of active sweeps to starting point."""
        self._scpi_mock.ask.return_value = "0,\"No error\""
        self.instr.reset_sweep()

        self._scpi_mock.write.assert_called_once_with("SOUR:SWE:RES:ALL")

    def test_reset_list_mode(self):
        """Test resetting of active sweeps to starting point."""
        self._scpi_mock.ask.return_value = "0,\"No error\""
        self.instr.reset_list_mode()

        self._scpi_mock.write.assert_called_once_with("SOUR:LIST:RES")

    def test_learn_list(self):
        """Test learning of the list."""
        self._scpi_mock.ask.return_value = "0,\"No error\""
        self.instr.learn_list()

        self._scpi_mock.write.assert_called_once_with("SOUR:LIST:LEAR")

    def test_reset_list(self):
        """Test learning of the list."""
        self._scpi_mock.ask.return_value = "0,\"No error\""
        self.instr.reset_list()

        self._scpi_mock.write.assert_called_once_with("ABOR:LIST")

    def test_load_fplist(self):
        """Test loading of frequency and power list."""
        self._scpi_mock.ask.return_value = "0,\"No error\""
        actual_start_freq = 1.0
        actual_stop_freq = 10.0
        actual_num_steps = 10
        actual_unit_freq = "Hz"
        actual_start_pow = 11.0
        actual_stop_pow = 20.0
        actual_unit_pow = "dB"
        actual_flist = "1.0Hz,2.0Hz,3.0Hz,4.0Hz,5.0Hz,6.0Hz,7.0Hz,8.0Hz,9.0Hz,10.0Hz"
        actual_plist = "11.0dB,12.0dB,13.0dB,14.0dB,15.0dB,16.0dB,17.0dB,18.0dB,19.0dB,20.0dB"    
        expected_calls = [
            call(f"SOUR:LIST:SEL 'list_{actual_start_freq}_{actual_stop_freq}_{actual_num_steps}'"),
            call('SOUR:LIST:FREQ ' + actual_flist),
            call('SOUR:LIST:POW ' + actual_plist)
        ]

        self.instr.load_fplist(
            fstart=actual_start_freq, fstop=actual_stop_freq, funit=actual_unit_freq,
            number_of_steps=actual_num_steps, pstart=actual_start_pow, pstop=actual_stop_pow, punit=actual_unit_pow
        )

        self._scpi_mock.write.assert_has_calls(expected_calls)


if __name__ == '__main__':
    unittest.main()



import logging
import os
import unittest
from unittest.mock import Mock, patch

import numpy as np

import sys
import tests.instruments.adwin.adwin_stub
sys.modules['ADwin'] = tests.instruments.adwin.adwin_stub
patcher = patch("ADwin.ADwin")
_patcher = patcher.start()

from qmi.core.exceptions import QMI_InstrumentException, QMI_TimeoutException
from qmi.core.context import QMI_Context
from qmi.instruments.adwin import adwin
from qmi.instruments.adwin import JagerMessTechnik_AdwinGoldII, JagerMessTechnik_AdwinProII

# Class constants from ADwin.ADwin class
ADWIN_DATATYPE_INT8 = 1
ADWIN_DATATYPE_INT16 = 2
ADWIN_DATATYPE_INT32 = 3
ADWIN_DATATYPE_SINGLE = 5
ADWIN_DATATYPE_DOUBLE = 6
ADWIN_DATATYPE_INT64 = 7


class AdwinProIITest(unittest.TestCase):
    def test_class_constants_are_ok(self) -> None:
        self._adwin = JagerMessTechnik_AdwinProII(QMI_Context("test"), "Adwin", 1234)
        self.assertEqual(self._adwin.MAX_PAR, 80)
        self.assertEqual(self._adwin.MAX_DATA, 200)
        self.assertEqual(self._adwin.MAX_PROCESS_NO, 10)


class AdwinGoldIITest(unittest.TestCase):
    def test_class_constants_are_ok(self) -> None:
        self._adwin = JagerMessTechnik_AdwinGoldII(QMI_Context("test"), "Adwin", 1234)
        self.assertEqual(self._adwin.MAX_PAR, 80)
        self.assertEqual(self._adwin.MAX_DATA, 200)
        self.assertEqual(self._adwin.MAX_PROCESS_NO, 10)


class ADwinTestBaseClass(unittest.TestCase):
    def setUp(self) -> None:
        self._adwin = adwin.Adwin_Base(QMI_Context("test"), "Adwin", 1234)
        self._adwin.MAX_PROCESS_NO = 1
        self._adwin.MAX_PAR = 2
        self._adwin.MAX_DATA = 3
        self._adwin._adwin.ADWIN_DATATYPE_INT8 = ADWIN_DATATYPE_INT8
        self._adwin._adwin.ADWIN_DATATYPE_DOUBLE = ADWIN_DATATYPE_DOUBLE
        self._adwin.open()

    def tearDown(self) -> None:
        logging.getLogger("qmi.instruments.adwin.adwin").setLevel(logging.NOTSET)

    def test_check_ready(self):
        """ This test should run without raising any exceptions """
        self._adwin._adwin.Test_Version = Mock(return_value=0)
        self._adwin.check_ready()

        self._adwin._adwin.Test_Version.assert_called_once()

    def test_check_ready_wrong_version(self):
        """ This test raises QMI_InstrumentException due to error in Test_Version """
        self._adwin._adwin.Test_Version = Mock(return_value="")
        with self.assertRaises(QMI_InstrumentException):
            self._adwin.check_ready()

    def test_get_processor_type(self):
        """ Get the processor type"""
        expected = "T11"
        self._adwin._adwin.Processor_Type = Mock(return_value=expected)
        processor_type = self._adwin.get_processor_type()
        self.assertEqual(processor_type, expected)
        self._adwin._adwin.Processor_Type.assert_called_once()

    def test_reboot_works_fine(self):
        """ Test reboot function."""
        self._adwin._adwin.Test_Version = Mock(return_value=0)
        self._adwin._adwin.Processor_Type = Mock(return_value="T12.1")
        self._adwin._adwin.ADwindir = os.path.split(os.path.abspath(adwin.__file__))[0]
        self._adwin._boot_file = "adwin.py"
        with patch("os.path.isfile", return_value=True) as p:
            p.start()
            self._adwin.reboot()

        p.clean()
        self._adwin._adwin.Boot.assert_called_once()
        self._adwin._adwin.Processor_Type.assert_called_once()
        self._adwin._adwin.Test_Version.assert_called_once()

    def test_reboot_exception(self):
        """ Test reboot function. It raises an QMI_InstrumentException due to not mocking Test_Version """
        self._adwin._adwin.Processor_Type = Mock(return_value="T12")
        self._adwin._adwin.ADwindir = os.path.split(os.path.abspath(adwin.__file__))[0]
        self._adwin._boot_file = "adwin.py"
        with self.assertRaises(QMI_InstrumentException):
            self._adwin.reboot()

        self._adwin._adwin.Processor_Type.assert_called_once()

    def test_reboot_wrong_file_path(self):
        """ Test reboot function. It raises an QMI_InstrumentException due to not mocking Test_Version """
        self._adwin._adwin.ADwindir = ""
        self._adwin._boot_file = "adwin.py"
        with self.assertRaises(QMI_InstrumentException):
            self._adwin.reboot()

    def test_get_workload(self):
        """ Test get_workload function. """
        expected = 0
        self._adwin._adwin.Workload = Mock(return_value=expected)
        ans = self._adwin.get_workload()
        self.assertEqual(ans, expected)

    def test_load_process(self):
        """ Test load_process function. """
        self._adwin._adwin.Load_Process = Mock()
        self._adwin.load_process("bin_file")
        self._adwin._adwin.Load_Process.assert_called_once()

    def test_start_process(self):
        """ Test start_process function. """
        self._adwin._adwin.Process_Status = Mock(return_value=0)
        self._adwin._adwin.Start_Process = Mock()
        self._adwin.start_process(1)
        self._adwin._adwin.Process_Status.assert_called_once()
        self._adwin._adwin.Start_Process.assert_called_once()

    def test_start_process_number_out_of_range(self):
        """ Test start_process function with wrong process number inputs """
        numbers = [0, 2]
        for number in numbers:
            with self.assertRaises(ValueError):
                self._adwin.start_process(number)

    def test_start_process_raises_exception_status(self):
        """ Test start_process function raises exception. """
        self._adwin._adwin.Process_Status = Mock(return_value=1)
        with self.assertRaises(QMI_InstrumentException):
            self._adwin.start_process(1)

    def test_stop_process(self):
        """ Test stop_process functions """
        self._adwin._adwin.Stop_Process = Mock(return_value=0)
        self._adwin.stop_process(1)
        self._adwin._adwin.Stop_Process.assert_called_once()

    def test_stop_process_number_out_of_range(self):
        """ Test stop_process function with wrong process number inputs """
        numbers = [0, 2]
        for number in numbers:
            with self.assertRaises(ValueError):
                self._adwin.stop_process(number)

    def test_wait_for_process(self):
        """ Test wait_for_process function """
        expected = 2
        number, timeout = 1, 0.1
        self._adwin._adwin.Process_Status = Mock()
        self._adwin._adwin.Process_Status.side_effect = [1, 0]  # has to wait once
        self._adwin.wait_for_process(number, timeout)
        self._adwin._adwin.Process_Status.assert_called_with(1)
        self.assertEqual(self._adwin._adwin.Process_Status.call_count, expected)

    def test_wait_for_process_number_out_of_range(self):
        """ Test wait_for_process function """
        numbers = [0, 2]
        for number in numbers:
            with self.assertRaises(ValueError):
                self._adwin.wait_for_process(number, 0.0)

    def test_wait_for_process_timeout_exception(self):
        """ Test wait_for_process function """
        number, timeout = 1, 0.01
        self._adwin._adwin.Process_Status = Mock()
        self._adwin._adwin.Process_Status.side_effect = [1, 1, 0]  # has to wait twice
        with self.assertRaises(QMI_TimeoutException):
            self._adwin.wait_for_process(number, timeout)

    def test_is_process_running_true(self):
        """ Test is_process_running function when process is running """
        self._adwin._adwin.Process_Status = Mock(return_value=1)
        ans = self._adwin.is_process_running(1)
        self.assertTrue(ans)

    def test_is_process_running_false(self):
        """ Test is_process_running function when process is not running """
        self._adwin._adwin.Process_Status = Mock(return_value=0)
        ans = self._adwin.is_process_running(1)
        self.assertFalse(ans)

    def test_is_process_running_wrong_slot_index(self):
        """ Test is_process_running with invalid slot index number """
        with self.assertRaises(ValueError):
            self._adwin.is_process_running(0)

    def test_get_par(self):
        """ Test get_par function """
        expected = 1
        self._adwin._adwin.Get_Par = Mock(return_value=expected)
        ans = self._adwin.get_par(1)
        self.assertEqual(ans, expected)

    def test_get_par_index_out_of_range(self):
        """ Test get_par function excepts when input index is out of range """
        indexes = [0, 3]
        for index in indexes:
            with self.assertRaises(ValueError):
                self._adwin.get_par(index)

    def test_get_fpar(self):
        """ Test get_fpar function """
        expected = 1.0
        self._adwin._adwin.Get_FPar_Double = Mock(return_value=expected)
        ans = self._adwin.get_fpar(1)
        self.assertEqual(ans, expected)

    def test_get_fpar_index_out_of_range(self):
        """ Test get_fpar function excepts when input index is out of range """
        indexes = [0, 3]
        for index in indexes:
            with self.assertRaises(ValueError):
                self._adwin.get_fpar(index)

    def test_get_par_block(self):
        """ test get_par_block function """
        par_first, par_count = 1, 2
        expected = [1, 2]
        self._adwin._adwin.Get_Par_Block = Mock(return_value=expected)
        ans = self._adwin.get_par_block(par_first, par_count)
        for e in range(len(expected)):
            self.assertEqual(ans[e], expected[e])

        self.assertIsInstance(ans, np.ndarray)

    def test_get_par_block_wrong_first_index(self):
        """ test get_par_block excepts when parameter first index is invalid """
        par_firsts = [0, 3]
        par_count = 2
        for par_first in par_firsts:
            with self.assertRaises(ValueError):
                self._adwin.get_par_block(par_first, par_count)

    def test_get_par_block_count_too_large(self):
        """ test get_par_block excepts when parameter count is too large """
        par_first = 1
        par_count = 3
        with self.assertRaises(ValueError):
            self._adwin.get_fpar_block(par_first, par_count)

    def test_get_fpar_block(self):
        """ test get_fpar_block function """
        par_first, par_count = 1, 2
        expected = [1.0, 2.0]
        self._adwin._adwin.Get_FPar_Block_Double = Mock(return_value=expected)
        ans = self._adwin.get_fpar_block(par_first, par_count)
        for e in range(len(expected)):
            self.assertEqual(ans[e], expected[e])

        self.assertIsInstance(ans, np.ndarray)

    def test_get_fpar_block_wrong_first_index(self):
        """ test get_fpar_block excepts when parameter first index is invalid """
        par_firsts = [0, 3]
        par_count = 2
        for par_first in par_firsts:
            with self.assertRaises(ValueError):
                self._adwin.get_fpar_block(par_first, par_count)

    def test_get_fpar_block_count_too_large(self):
        """ test get_par_block excepts when parameter count is too large """
        par_first = 1
        par_count = 3
        with self.assertRaises(ValueError):
            self._adwin.get_fpar_block(par_first, par_count)

    def test_set_par(self):
        """ test set_par """
        par_idx, value = 1, 0
        self._adwin._adwin.Set_Par = Mock()
        self._adwin.set_par(par_idx, value)

        self._adwin._adwin.Set_Par.assert_called_once_with(par_idx, value)

    def test_set_par_index_out_of_range(self):
        """ test set_par raises exception when parameter index is out of range """
        par_indexes = [0, 3]
        value = 1
        for par_idx in par_indexes:
            with self.assertRaises(ValueError):
                self._adwin.set_par(par_idx, value)

    def test_set_par_value_not_int(self):
        """ test set_par raises exception when parameter value is not an integer """
        par_idx = 1
        values = ["3", 1.0, None]  # Note: booleans True and False do not raise a TypeError!
        for value in values:
            with self.assertRaises(TypeError):
                self._adwin.set_par(par_idx, value)

    def test_set_fpar(self):
        """ test set_fpar """
        par_idx, value = 1, 0.0
        self._adwin._adwin.Set_FPar_Double = Mock()
        self._adwin.set_fpar(par_idx, value)

        self._adwin._adwin.Set_FPar_Double.assert_called_once_with(par_idx, value)

    def test_set_fpar_index_out_of_range(self):
        """ test set_fpar raises exception when parameter index is out of range """
        par_indexes = [0, 3]
        value = 1.0
        for par_idx in par_indexes:
            with self.assertRaises(ValueError):
                self._adwin.set_fpar(par_idx, value)

    def test_set_fpar_value_not_int(self):
        """ test set_fpar raises exception when parameter value is not an integer nor a float """
        par_idx = 1
        values = ["3", None]  # Note: booleans True and False do not raise a TypeError!
        for value in values:
            with self.assertRaises(TypeError):
                self._adwin.set_fpar(par_idx, value)

    def test_get_data_length(self):
        """ test get_data_length function """
        expected = 2
        par_idx = 3
        self._adwin._adwin.Data_Length = Mock(return_value=expected)
        ans = self._adwin.get_data_length(par_idx)
        self.assertEqual(ans, expected)

    def test_get_data_length_data_index_out_of_range(self):
        """ test get_data_length function raises an exception when data index is out of range """
        par_indexes = [0, 4]
        for par_idx in par_indexes:
            with self.assertRaises(ValueError):
                self._adwin.get_data_length(par_idx)

    def test_get_data_long(self):
        """ test get_data function returning a long array """
        expected = [3, 4]
        data_idx, first_index, count = 1, 1, 2
        self._adwin._adwin.Data_Type = Mock(return_value=[ADWIN_DATATYPE_INT8, "byte"])
        self._adwin._adwin.GetData_Long = Mock(return_value=expected)
        ans = self._adwin.get_data(data_idx, first_index, count)
        for e in range(len(expected)):
            self.assertEqual(ans[e], expected[e])

    def test_get_data_double(self):
        """ test get_data function returning a double array """
        expected = [3.0, 4.0]
        data_idx, first_index, count = 1, 1, 2
        self._adwin._adwin.Data_Type = Mock(return_value=[ADWIN_DATATYPE_DOUBLE, "float64"])
        self._adwin._adwin.GetData_Double = Mock(return_value=expected)
        ans = self._adwin.get_data(data_idx, first_index, count)
        for e in range(len(expected)):
            self.assertEqual(ans[e], expected[e])

    def test_get_data_data_index_out_of_range(self):
        """ test get_data function raises an exception when data index is out of range """
        data_indexes = [0, 4]
        first_index, count = 1, 2
        for data_idx in data_indexes:
            with self.assertRaises(ValueError):
                self._adwin.get_data(data_idx, first_index, count)

    def test_get_data_first_index_too_small(self):
        """ test get_data function raises an exception when first index is too small """
        data_idx, first_index, count = 1, 0, 2
        with self.assertRaises(ValueError):
            self._adwin.get_data(data_idx, first_index, count)

    def test_get_data_count_too_small(self):
        """ test get_data function raises an exception when count is too small """
        data_idx, first_index, count = 1, 1, 0
        with self.assertRaises(ValueError):
            self._adwin.get_data(data_idx, first_index, count)

    def test_get_data_invalid_index_type(self):
        """ See that QMI_InstrumentException gets raised with invalid index data type"""
        data_idx, first_index, count = 1, 1, 2
        self._adwin._adwin.Data_Type = Mock(return_value=[ADWIN_DATATYPE_DOUBLE, "str"])
        with self.assertRaises(QMI_InstrumentException):
            self._adwin.get_data(data_idx, first_index, count)

    def test_get_full_data_array_long(self):
        """ test get_full_data function returning a long array """
        data_idx = 3
        expected = [3, 4]
        self._adwin._adwin.Data_Type = Mock(return_value=[ADWIN_DATATYPE_INT8, "byte"])
        self._adwin._adwin.GetData_Long = Mock(return_value=expected)
        ans = self._adwin.get_full_data(data_idx)
        for e in range(len(expected)):
            self.assertEqual(ans[e], expected[e])

    def test_get_full_data_array_double(self):
        """ test get_full_data function returning a double array """
        data_idx = 3
        expected = [3.0, 4.0]
        self._adwin._adwin.Data_Type = Mock(return_value=[ADWIN_DATATYPE_DOUBLE, "float64"])
        self._adwin._adwin.GetData_Long = Mock(return_value=expected)
        ans = self._adwin.get_full_data(data_idx)
        for e in range(len(expected)):
            self.assertEqual(ans[e], expected[e])

    def test_get_full_data_data_index_out_of_range(self):
        """ test get_full_data function raises an exception when data index is out of range """
        data_indexes = [0, 4]
        for data_idx in data_indexes:
            with self.assertRaises(ValueError):
                self._adwin.get_full_data(data_idx)

    def test_set_data_long_list(self):
        """ test set_data function with a list of integers """
        data_idx, first_index, value = 3, 1, [1, 2, 3]
        self._adwin._adwin.Data_Type = Mock(return_value=[ADWIN_DATATYPE_INT16, "short"])
        self._adwin._adwin.SetData_Long = Mock()
        self._adwin.set_data(data_idx, first_index, value)

        self._adwin._adwin.Data_Type.assert_called_once()
        self._adwin._adwin.SetData_Long.assert_called_once()

    def test_set_data_float_list(self):
        """ test set_data function with a list of floats """
        data_idx, first_index, value = 3, 1, [1.0, 2.0, 3.0]
        self._adwin._adwin.Data_Type = Mock(return_value=[ADWIN_DATATYPE_SINGLE, "float32"])
        self._adwin._adwin.SetData_Double = Mock()
        self._adwin.set_data(data_idx, first_index, value)

        self._adwin._adwin.Data_Type.assert_called_once()
        self._adwin._adwin.SetData_Double.assert_called_once()

    def test_set_data_long_array(self):
        """ test set_data function with an array of integers """
        data_idx, first_index, value = 3, 1, np.array([1, 2, 3])
        self._adwin._adwin.Data_Type = Mock(return_value=[ADWIN_DATATYPE_INT32, "int32"])
        self._adwin._adwin.SetData_Long = Mock()
        self._adwin.set_data(data_idx, first_index, value)

        self._adwin._adwin.Data_Type.assert_called_once()
        self._adwin._adwin.SetData_Long.assert_called_once()

    def test_set_data_float_array(self):
        """ test set_data function with an array of floats """
        data_idx, first_index, value = 3, 1, np.array([1.0, 2.0, 3.0])
        self._adwin._adwin.Data_Type = Mock(return_value=[ADWIN_DATATYPE_SINGLE, "float"])
        self._adwin._adwin.SetData_Double = Mock()
        self._adwin.set_data(data_idx, first_index, value)

        self._adwin._adwin.Data_Type.assert_called_once()
        self._adwin._adwin.SetData_Double.assert_called_once()

    def test_set_data_int_array_with_invalid_value(self):
        """ test set_data function with an array of integers """
        data_idx, first_index, value = 3, 1, np.array([1, 2, None])
        self._adwin._adwin.Data_Type = Mock(return_value=[ADWIN_DATATYPE_INT64, "int64"])
        self._adwin._adwin.SetData_Long = Mock()
        with self.assertRaises(ValueError):
            self._adwin.set_data(data_idx, first_index, value)

    def test_set_data_wrong_index(self):
        """ test set_data with wrong index number"""
        with self.assertRaises(ValueError):
            self._adwin.set_data(0, 1, np.ndarray([2]))

    def test_set_data_first_index_zero(self):
        """ test set_data with wrong first index number"""
        with self.assertRaises(ValueError):
            self._adwin.set_data(1, 0, np.ndarray([2]))

    def test_get_fifo_filled(self):
        """ Test function get_fifo_filled """
        data_idx = 3
        expected = 109
        self._adwin._adwin.Fifo_Full = Mock(return_value=expected)
        ans = self._adwin.get_fifo_filled(data_idx)

        self.assertEqual(ans, expected)

    def test_get_fifo_filled_data_index_out_of_range(self):
        """ Test function get_fifo_filled raises an exception when data index is out of range """
        data_indexes = [0, 4]
        for data_idx in data_indexes:
            with self.assertRaises(ValueError):
                self._adwin.get_fifo_filled(data_idx)

    def test_get_fifo_room(self):
        """ Test function get_fifo_room """
        data_idx = 3
        expected = 91
        self._adwin._adwin.Fifo_Empty = Mock(return_value=expected)
        ans = self._adwin.get_fifo_room(data_idx)

        self.assertEqual(ans, expected)

    def test_get_fifo_room_data_index_out_of_range(self):
        """ Test function get_fifo_room raises an exception when data index is out of range """
        data_indexes = [0, 4]
        for data_idx in data_indexes:
            with self.assertRaises(ValueError):
                self._adwin.get_fifo_room(data_idx)

    def test_read_fifo_long(self):
        """ test read_fifo function returning a long array """
        expected = [3, 4]
        data_idx, count = 1, 2
        self._adwin._adwin.Data_Type = Mock(return_value=[ADWIN_DATATYPE_INT64, "int64"])
        self._adwin._adwin.GetFifo_Long = Mock(return_value=expected)
        ans = self._adwin.read_fifo(data_idx, count)
        for e in range(len(expected)):
            self.assertEqual(ans[e], expected[e])

    def test_read_fifo_double(self):
        """ test read_fifo function returning a double array """
        expected = [3.0, 4.0]
        data_idx, count = 1, 2
        self._adwin._adwin.Data_Type = Mock(return_value=[ADWIN_DATATYPE_DOUBLE, "float64"])
        self._adwin._adwin.GetFifo_Double = Mock(return_value=expected)
        ans = self._adwin.read_fifo(data_idx, count)
        for e in range(len(expected)):
            self.assertEqual(ans[e], expected[e])

    def test_read_fifo_data_index_out_of_range(self):
        """ test read_fifo function raises an exception when data index is out of range """
        data_indexes = [0, 4]
        count = 2
        for data_idx in data_indexes:
            with self.assertRaises(ValueError):
                self._adwin.read_fifo(data_idx, count)

    def test_write_fifo_long_list(self):
        """ test write_fifo function with a list of integers """
        data_idx, value = 3, [1, 2, 3]
        self._adwin._adwin.Data_Type = Mock(return_value=[ADWIN_DATATYPE_INT16, "short"])
        self._adwin._adwin.SetFifo_Long = Mock()
        self._adwin.write_fifo(data_idx, value)

        self._adwin._adwin.Data_Type.assert_called_once()
        self._adwin._adwin.SetFifo_Long.assert_called_once()

    def test_write_fifo_float_list(self):
        """ test write_fifo function with a list of floats """
        data_idx, value = 3, [1.0, 2.0, 3.0]
        self._adwin._adwin.Data_Type = Mock(return_value=[ADWIN_DATATYPE_SINGLE, "float32"])
        self._adwin._adwin.SetFifo_Double = Mock()
        self._adwin.write_fifo(data_idx, value)

        self._adwin._adwin.Data_Type.assert_called_once()
        self._adwin._adwin.SetFifo_Double.assert_called_once()

    def test_write_fifo_long_array(self):
        """ test write_fifo function with an array of integers """
        data_idx, value = 3, np.array([1, 2, 3])
        self._adwin._adwin.Data_Type = Mock(return_value=[ADWIN_DATATYPE_INT64, "int64"])
        self._adwin._adwin.SetFifo_Long = Mock()
        self._adwin.write_fifo(data_idx, value)

        self._adwin._adwin.Data_Type.assert_called_once()
        self._adwin._adwin.SetFifo_Long.assert_called_once()

    def test_write_fifo_float_array(self):
        """ test write_fifo function with an array of floats """
        data_idx, value = 3, np.array([1.0, 2.0, 3.0])
        self._adwin._adwin.Data_Type = Mock(return_value=[ADWIN_DATATYPE_DOUBLE, "float64"])
        self._adwin._adwin.SetFifo_Double = Mock()
        self._adwin.write_fifo(data_idx, value)

        self._adwin._adwin.Data_Type.assert_called_once()
        self._adwin._adwin.SetFifo_Double.assert_called_once()

    def test_write_fifo_int_array_with_invalid_value(self):
        """ test write_fifo function with an array of integers """
        data_idx, value = 3, np.array([1, 2, "tripel"])
        self._adwin._adwin.Data_Type = Mock(return_value=[ADWIN_DATATYPE_INT32, "long"])
        self._adwin._adwin.SetFifo_Long = Mock()
        with self.assertRaises(ValueError):
            self._adwin.write_fifo(data_idx, value)

    def test_write_fifo_invalid_index(self):
        """ test write_fifo function with wrong index number"""
        with self.assertRaises(ValueError):
            self._adwin.write_fifo(201, np.array([3]))

    @patch('sys.platform', 'linux1')
    def test_set_file_to_data_linux(self):
        """ test set_file_to_data function in patched linux environment """
        expected_adwin_data_type = 3
        data_idx, first_index, file_path = 3, 1, os.path.abspath(__file__)
        self._adwin._adwin.Data_Type = Mock(return_value=[ADWIN_DATATYPE_INT32, "long"])
        self._adwin._adwin.File2Data = Mock()
        self._adwin.set_file_to_data(data_idx, first_index, file_path)

        self._adwin._adwin.Data_Type.assert_called_once()
        self._adwin._adwin.File2Data.assert_called_once_with(file_path, expected_adwin_data_type, data_idx, first_index)

    @patch('sys.platform', 'win32')
    def test_set_file_to_data_windows(self):
        """ test set_file_to_data function in patched windows environment """
        expected_adwin_data_type = 2
        data_idx, first_index, file_path = 3, 1, os.path.abspath(__file__)
        self._adwin._adwin.Data_Type = Mock(return_value=[ADWIN_DATATYPE_INT32, "int32"])
        self._adwin._adwin.File2Data = Mock()
        self._adwin.set_file_to_data(data_idx, first_index, file_path)

        self._adwin._adwin.Data_Type.assert_called_once()
        self._adwin._adwin.File2Data.assert_called_once_with(file_path, expected_adwin_data_type, data_idx, first_index)

    def test_set_file_to_data_index_out_of_range(self):
        """ test set_file_to_data function raises an exception when data index is out of range """
        data_indexes = [0, 4]
        first_index, file_path = 1, os.path.abspath(__file__)
        for data_idx in data_indexes:
            with self.assertRaises(ValueError):
                self._adwin.set_file_to_data(data_idx, first_index, file_path)

    def test_set_file_to_data_first_index_too_small(self):
        """ test set_file_to_data function raises an exception when data index is out of range """
        data_indexes = [1, 2]
        first_index, file_path = 0, os.path.abspath(__file__)
        for data_idx in data_indexes:
            with self.assertRaises(ValueError):
                self._adwin.set_file_to_data(data_idx, first_index, file_path)

    def test_set_file_to_data_wrong_index_type(self):
        """ test set_file_to_data function raises an exception when data index type is not an integer """
        # Suppress logging.
        logging.getLogger("qmi.instruments.adwin.adwin").setLevel(logging.CRITICAL)
        expected = "QMI_InstrumentException(\'Target data array should be of type long.\')"
        self._adwin._adwin.Data_Type = Mock(return_value=[ADWIN_DATATYPE_DOUBLE, "float32"])
        data_idx = 3.0
        first_index, file_path = 1, os.path.abspath(__file__)
        with self.assertRaises(QMI_InstrumentException) as exc:
            self._adwin.set_file_to_data(data_idx, first_index, file_path)

        self.assertEqual(repr(exc.exception), expected)

    def test_set_file_to_data_file_is_missing(self):
        """ test set_file_to_data function raises an exception when the file specified is missing """
        # Suppress logging.
        logging.getLogger("qmi.instruments.adwin.adwin").setLevel(logging.CRITICAL)
        expected = "QMI_InstrumentException(\'The specified file does not exist.\')"
        self._adwin._adwin.Data_Type = Mock(return_value=[ADWIN_DATATYPE_INT16, "short"])
        data_idx, first_index, file_path = 3, 1, "niet_bestaand.bestand"
        with self.assertRaises(QMI_InstrumentException) as exc:
            self._adwin.set_file_to_data(data_idx, first_index, file_path)

        self.assertEqual(repr(exc.exception), expected)


if __name__ == "__main__":
    unittest.main()



""" QMI instrument driver for the PicoQuant PicoHarp 300 instrument.

The instrument driver makes use of the manufacturer provided software libraries, "phlib.so" for Linux OS,
or "phlib.dll" or "phlib64.dll" for 32-bit and 64-bit Windows OS, respectively.
Please find the licence terms for these files in the dedicated software package for the PicoHarp instrument at
https://www.picoquant.com/products/category/tcspc-and-time-tagging-modules/picoharp-300-stand-alone-tcspc-module-with-usb-interface
--> "Software" tab --> download link in "Current software and developer's library version".
"""
import ctypes
import enum
import logging
from typing import List

from qmi.core.exceptions import QMI_InvalidOperationException
from qmi.core.rpc import rpc_method
from qmi.instruments.picoquant.support._library_wrapper import _LibWrapper
from qmi.instruments.picoquant._picoquant import _str_to_enum, _PicoquantHarp
from qmi.instruments.picoquant.support._events import _MODE

_logger = logging.getLogger(__name__)


@enum.unique
class _FLAG(enum.IntFlag):
    """Bitfield constants for the return value of the :func:`~PicoHarpDevice.getFlags` function.

    These are defined as preprocessor symbols in the ``phdefin.h`` C header file.
    """
    OVERFLOW = 0x0040
    """Histogram mode only. It indicates that a histogram measurement has reached the maximum count 
    as specified via PH_SetStopOverflow."""
    FIFOFULL = 0x0003
    """TTTR mode only. It indicates that the data FiFo has run full. The measurement will then have
    to be aborted as data integrity is no longer maintained."""
    SYSERROR = 0x0100
    """Indicates an error of the hardware or internal software. The user should in this case
    call the library routine PH_GetHardwareDebugInfo and provide the result to PicoQuant support."""


@enum.unique
class _WARNING(enum.IntFlag):
    """Bitfield constants for the return value of the :func:`~PicoHarpDevice.getWarnings` function.

    These are defined as preprocessor symbols in the ``phdefin.h`` C header file. For full descriptions
    see the chapter '11. Warnings' from the PicoHarp PHLib Manual.
    """
    INP0_RATE_ZERO = 0x0001
    """No counts are detected at input channel 0."""
    INP0_RATE_TOO_LOW = 0x0002
    """The count rate at input channel 0 is below 1 kHz and the sync divider is >1."""
    INP0_RATE_TOO_HIGH = 0x0004
    """You have selected T2 mode and the count rate at input channel 0 is higher than 5 MHz.
    The measurement will inevitably lead to a FiFo overrun."""
    INP1_RATE_ZERO = 0x0010
    """No counts are detected at input channel 1."""
    INP1_RATE_TOO_HIGH = 0x0040
    """If you have selected T2 mode then this warning means the count rate at input channel 1 is higher than 5 MHz. 
    The measurement will inevitably lead to a FiFo overrun.
    In histogramming and T3 mode this warning is issued when the input rate is >10 MHz. This will probably lead to 
    deadtime artefacts."""
    INP_RATE_RATIO = 0x0100
    """This warning is issued in histogramming and T3 mode when the rate at input 1 is over 5% of the rate
    at input 0."""
    DIVIDER_GREATER_ONE = 0x0200
    """You have selected T2 mode and the sync divider is set larger than 1. This is probably not intended."""
    TIME_SPAN_TOO_SMALL = 0x0400
    """This warning is issued in histogramming and T3 mode when the sync period (1/Rate[ch0]) is longer than 
    the start to stop time span that can be covered by the histogram or by the T3 mode records."""
    OFFSET_UNNECESSARY = 0x0800
    """This warning is issued in histogramming and T3 mode when an offset >0 is set even though the sync period
    (1/Rate0) can be covered by the measurement time span without using an offset. The offset may lead to events
    getting discarded."""


class PicoQuant_PicoHarp300(_PicoquantHarp):
    """Instrument driver for the PicoQuant PicoHarp 300."""

    _MODEL = "PH"
    # From phdefin.h
    _MAXDEVNUM = 8
    _TTREADMAX = 131072

    @property
    def _max_dev_num(self):
        return self._MAXDEVNUM

    @property
    def _ttreadmax(self):
        return self._TTREADMAX

    @property
    def _model(self):
        return self._MODEL

    @property
    def _lib(self) -> _LibWrapper:
        if self._lazy_lib is None:  # type: ignore
            self._lazy_lib = _LibWrapper('PH')
        return self._lazy_lib

    @rpc_method
    def initialize(self, mode_str: str) -> None:
        """Initialize the device.

        This routine must be called before any of the other methods, except OpenDevice, CloseDevice,
        GetErrorString and GetLibraryVersion can be used.

        Arguments:
            mode_str (str): Opening mode. Can be any of 'HIST', 'T2', 'T3'.
                            The latest driver version V3.0+ supports T2 and T3 modes.

        Raises:
            ValueError: if either argument is an unknown string value.
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()

        if self._measurement_running:
            raise QMI_InvalidOperationException("Measurement still active")
        mode = _str_to_enum(_MODE, mode_str)
        if mode not in (_MODE.HIST, _MODE.T2, _MODE.T3):
            raise ValueError("Unsupported measurement mode")

        if self._lib_version.startswith("2") and mode is _MODE.T3:
            raise ValueError("Unsupported measurement mode")

        with self._device_lock:
            self._lib.Initialize(self._devidx, mode.value)
            self._mode = mode

        # Reset the event filter configuration.
        self.set_event_filter(reset_filter=True)
        self.set_block_events(False)
        _logger.info("[%s] Initialized device with mode %s", self._name, mode.value)

    @rpc_method
    def set_sync_offset(self, sync_offset: int) -> None:
        self._check_is_open()
        with self._device_lock:
            self._lib.SetSyncOffset(self._devidx, sync_offset)

    @rpc_method
    def get_flags(self) -> List[str]:
        self._check_is_open()
        with self._device_lock:
            bitset = ctypes.c_int()
            self._lib.GetFlags(self._devidx, bitset)
            flags = _FLAG(bitset.value)
            return [flag.name for flag in _FLAG if flag in flags and flag.name is not None]

    @rpc_method
    def get_warnings(self) -> List[str]:
        self._check_is_open()
        with self._device_lock:
            bitset = ctypes.c_int()
            self._lib.GetWarnings(self._devidx, bitset)
            warnings = _WARNING(bitset.value)
            return [warning.name for warning in _WARNING if warning in warnings and warning.name is not None]

    @rpc_method
    def calibrate(self) -> None:
        self._check_is_open()
        with self._device_lock:
            self._lib.Calibrate(self._devidx)

    @rpc_method
    def set_input_cfd(self, channel: int, level: int, zc: int) -> None:
        """Set Constant Fraction Discriminator (CFD) levels for input channel.

        Notes:
            This method can only be called on the TimeHarp 260 P model.

        Args:
            channel: Channel index (range from 0 to `number_of_channels` - 1).
            level: CFD discriminator level in millivolts (mV) range 0 to -1200 mV.
            zc: CFD zero cross level in millivolts (mV). range 0 to -40 mV.
        """
        self._check_is_open()
        with self._device_lock:
            self._lib.SetInputCFD(self._devidx, channel, level, zc)



import os.path
import tempfile
import time
import unittest

import numpy as np
import h5py

from qmi.data.hdf5recorder import HDF5Recorder


class HDF5RecorderTestCase(unittest.TestCase):

    def test_smoke_test(self):
        dtype1 = np.dtype([("timestamp", np.float64), ("value1", np.float64)])
        dtype2 = np.dtype([("timestamp", np.float64), ("value1", np.float64), ("value2", np.float64)])

        temp_file = tempfile.mktemp()

        hdf5 = HDF5Recorder(temp_file)

        ds1 = []
        ds2 = []
        for i in range(3000):
            timestamp = time.time()
            v1 = np.random.randn()
            v2 = np.random.randn()
            ds1.append(np.array((timestamp, v1), dtype=dtype1))
            ds2.append(np.array((timestamp, v1, v2), dtype=dtype2))

            if len(ds1) == 97:
                hdf5.record("g1/g2/dataset1", ds1)
                ds1 = []

            if len(ds2) == 197:
                hdf5.record("dataset2", ds2)
                ds2 = []

            time.sleep(0.001)

        hdf5.record("g1/g2/dataset1", ds1)
        hdf5.record("dataset2", ds2)

        hdf5.close()


class TestRecorder(unittest.TestCase):

    def setUp(self):
        self._dir = tempfile.TemporaryDirectory()

    def tearDown(self):
        self._dir.cleanup()

    def test_record(self):

        np.random.seed(112223)

        n1 = 5000
        dtype1 = np.dtype([('x', np.uint32), ('y', np.float64)])
        test_data1 = np.empty(n1, dtype=dtype1)
        test_data1["x"] = np.random.randint(0, 2**32-1, n1, dtype=np.uint32)
        test_data1["y"] = np.random.uniform(0, 10.0, n1)

        n2 = 15000
        dtype2 = np.dtype([('p', np.int64), ('q', np.uint8)])
        test_data2 = np.empty(n2, dtype=dtype2)
        test_data2["p"] = np.random.randint(1-2**63, 2**63-1, n2, dtype=np.int64)
        test_data2["q"] = np.random.randint(0, 2**8-1, n2, dtype=np.uint8)

        file_name = os.path.join(self._dir.name, "test_record.h5")
        rec = HDF5Recorder(file_name, write_interval=1.0)

        p1 = 0
        p2 = 0
        for (q1, q2) in [(1234, 2345), (2345, 4444), (3000, 5000), (4010, 8000), (5000, 11000), (5000, 15000)]:
            rec.record("d1", test_data1[p1:q1])
            rec.record("d2", test_data2[p2:q2])
            p1 = q1
            p2 = q2
            time.sleep(0.8)

        rec.close()

        # Check HDF5 file contents.
        hdf = h5py.File(file_name, "r")
        try:
            self.assertEqual(set(hdf.keys()), {"d1", "d2"})
            self.assertEqual(hdf["d1"].shape, test_data1.shape)
            self.assertEqual(hdf["d1"].dtype, test_data1.dtype)
            self.assertTrue(np.all(hdf["d1"][:] == test_data1))
            self.assertEqual(hdf["d2"].shape, test_data2.shape)
            self.assertEqual(hdf["d2"].dtype, test_data2.dtype)
            self.assertTrue(np.all(hdf["d2"][:] == test_data2))
        finally:
            hdf.close()

    def test_attributes(self):

        np.random.seed(112224)

        test_data1 = np.arange(11)
        test_data2 = np.arange(22)

        file_name = os.path.join(self._dir.name, "test_attributes.h5")
        rec = HDF5Recorder(file_name, write_interval=1.0)

        # Write to dataset.
        rec.record("d1", test_data1)
        time.sleep(1.2)

        # Add attribute to existing dataset.
        rec.set_attribute("d1", "one", 1)
        rec.set_attribute("d1", "more", "MOAR")

        # Add attribute to non-existing datasets.
        rec.set_attribute("d2", "two", 2.0)
        rec.set_attribute("d3", "three", 3)
        time.sleep(1.2)

        # Overwrite existing attribute and add new attribute
        rec.set_attribute("d1", "more", "less")
        rec.set_attribute("d1", "neg", -1000)
        time.sleep(1.2)

        # Create dataset that already has pending attribute.
        rec.record("d2", test_data2)
        rec.close()

        # Check HDF5 file contents.
        hdf = h5py.File(file_name, "r")
        try:
            self.assertEqual(set(hdf.keys()), {"d1", "d2"})
            self.assertEqual(hdf["d1"].shape, test_data1.shape)
            self.assertEqual(hdf["d1"].dtype, test_data1.dtype)
            self.assertTrue(np.all(hdf["d1"][:] == test_data1))
            self.assertEqual(dict(hdf["d1"].attrs), {"one": 1, "more": "less", "neg": -1000})
            self.assertEqual(hdf["d2"].shape, test_data2.shape)
            self.assertEqual(hdf["d2"].dtype, test_data2.dtype)
            self.assertTrue(np.all(hdf["d2"][:] == test_data2))
            self.assertEqual(dict(hdf["d2"].attrs), {"two": 2.0})
        finally:
            hdf.close()


if __name__ == "__main__":
    unittest.main()



VI_ERROR_TMO = int(0xBFFF0015 - 0x100000000)

# correct
vendorid_1 = 0x0699
productid_1 = 0x3000
serialnr_1 = "XYZ"
# incorrect
vendorid_2 = "x0699"
productid_2 = "x3000"
serialnr_2 = "IJK"
# correct
vendorid_3 = 0xbebe
productid_3 = 0xcafe
serialnr_3 = "ABC"
# VISA-style resource strings
visa_str_1 = f"USB1::0x{vendorid_1:04x}::0x{productid_1:04x}::{serialnr_1}::INSTR"
visa_str_2 = f"USB2::{vendorid_2}::{productid_2}::{serialnr_2}::INSTR"
visa_str_3 = f"USBTMC::0x{vendorid_3:04x}::0x{productid_3:04x}::{serialnr_3}::INSTR"


class Error(Exception):
    """Abstract basic exception class for this module."""

    pass


class VisaIOError(Error):
    """Exception class for VISA I/O errors.

    Please note that all values for "error_code" are negative according to the
    specification (VPP-4.3.2, observation 3.3.2) and the NI implementation.

    """

    def __init__(self, error_code: int) -> None:
        abbreviation, description = ("VI_ERROR_TMO", "Timeout expired before operation " "completed.")
        super(VisaIOError, self).__init__(
            "%s (%d): %s" % (abbreviation, error_code, description)
        )
        self.error_code = error_code
        self.abbreviation = abbreviation
        self.description = description

    def __reduce__(self):
        """Store the error code when pickling."""
        return (VisaIOError, (self.error_code,))


class ResourceManager:
    RESOURCES = [visa_str_1, visa_str_2, visa_str_3]

    def open_resource(self, visa_resource):
        return visa_resource

    def list_resources(self):
        return self.RESOURCES


class errors:
    VisaIOError = VisaIOError
    VI_ERROR_TMO = VI_ERROR_TMO



#! /usr/bin/env python3

import time

import qmi
from qmi.instruments.dummy.noisy_sine_generator import NoisySineGenerator

qmi.start("nsg_server", "qmi.conf")

nsg = qmi.make_instrument("nsg", NoisySineGenerator)

time.sleep(0.100)
input("\nSimulated noisy sine generator instrument active, press Enter to quit ...\n")

qmi.stop()



"""" Wrapper for the external library component."""
import enum
import sys
from ctypes import CDLL

from qmi.instruments.high_finesse.support import wlmData, wlmConst

DLL_PATH_WINDOWS = "wlmData.dll"
DLL_PATH_LINUX = "libwlmData.so"
DLL_PATH_MACOS = "libwlmData.dylib"


@enum.unique
class WlmGetErr(enum.Enum):
    """Return error values of GetFrequency, GetWavelength, GetWLMVersion, GetPowerNum and GetOptionInfo."""
    NO_VALUE = wlmConst.ErrNoValue
    NO_SIGNAL = wlmConst.ErrNoSignal
    BAD_SIGNAL = wlmConst.ErrBadSignal
    LOW_SIGNAL = wlmConst.ErrLowSignal
    BIG_SIGNAL = wlmConst.ErrBigSignal
    WLM_MISSING = wlmConst.ErrWlmMissing
    NOT_AVAILABLE = wlmConst.ErrNotAvailable
    NOTHING_CHANGED = wlmConst.InfNothingChanged
    NO_PULSE = wlmConst.ErrNoPulse
    CHANNEL_NOT_AVAILABLE = wlmConst.ErrChannelNotAvailable
    DIV_ZERO = wlmConst.ErrDiv0
    OUT_OF_RANGE = wlmConst.ErrOutOfRange
    UNIT_NOT_AVAILABLE = wlmConst.ErrUnitNotAvailable


@enum.unique
class WlmTempErr(enum.Enum):
    """Return error values of GetTemperature and GetPressure."""
    TEMP_NOT_MEASURED = wlmConst.ErrTempNotMeasured
    TEMP_NOT_AVAILABLE = wlmConst.ErrTempNotAvailable
    TEMP_WLM_MISSING = wlmConst.ErrTempWlmMissing


@enum.unique
class WlmGainErr(enum.Enum):
    """Return error values of GetGain."""
    GAIN_NOT_AVAILABLE = wlmConst.ErrGainNotAvailable
    GAIN_WLM_MISSING = wlmConst.ErrGainWlmMissing
    GAIN_CHANNEL_NOT_AVAILABLE = wlmConst.ErrGainChannelNotAvailable
    GAIN_OUT_OF_RANGE = wlmConst.ErrGainOutOfRange
    GAIN_PARAMETER_OUT_OF_RANGE = wlmConst.ErrGainParameterOutOfRange


@enum.unique
class WlmMmiErr(enum.Enum):
    """Return error values of GetMultimodeInfo."""
    MMI_NOT_AVAILABLE = wlmConst.ErrMMINotAvailable
    MMI_WLM_MISSING = wlmConst.ErrMMIWlmMissing
    MMI_CHANNEL_NOT_AVAILABLE = wlmConst.ErrMMIChannelNotAvailable
    MMI_OUT_OF_RANGE = wlmConst.ErrMMIOutOfRange
    MMI_PARAMETER_OUT_OF_RANGE = wlmConst.ErrMMIParameterOutOfRange


@enum.unique
class WlmDistanceErr(enum.Enum):
    """Return error values of GetDistance."""
    DISTANCE_NOT_AVAILABLE = wlmConst.ErrDistanceNotAvailable
    DISTANCE_WLM_MISSING = wlmConst.ErrDistanceWlmMissing


class _LibWrapper:

    def __init__(self) -> None:
        """ A Library wrapper that hides the (platform-specific) HighFinesse Wavemeter libraries.

        Raises:
            OSError: Trying to run on an unsupported platform, e.g. "cygwin", or file not found.
        """
        if sys.platform.startswith("win"):
            wlmData.LoadDLL(DLL_PATH_WINDOWS)
        elif sys.platform.startswith("linux"):
            wlmData.LoadDLL(DLL_PATH_LINUX)
        elif sys.platform.startswith("darwin"):
            wlmData.LoadDLL(DLL_PATH_MACOS)
        else:
            raise OSError("Unsupported platform")

        self._dll = wlmData.dll

    @property
    def dll(self) -> CDLL:
        """ A reference to the actual dll."""
        assert self._dll is not None
        return self._dll



"""
Teraxion, narrowband tunable optical filters.

The qmi.instruments.teraxion package provides support for:
  - Teraxion TFN model (tested).
"""
from qmi.instruments.teraxion.tfn import Teraxion_TFN



import unittest
from unittest.mock import MagicMock, Mock, call, patch
import logging

import qmi.core.exceptions

import qmi.instruments
from tests.patcher import PatcherQmiContext as QMI_Context


# Mock import "mcculw"
class mcculw:
    ul = None
    enums = None
    structs = None


class TestFosUnix(unittest.TestCase):
    """ Unit tests for Bristol FOS. """

    UNIQUE_ID = 'DEADBEAF'

    @patch("sys.platform", "linux")
    def setUp(self):
        logging.getLogger("qmi.instruments.bristol.fos").setLevel(logging.CRITICAL)
        import qmi.instruments.bristol.fos

        # Substitute a mock object in place of the "uldaq" module.
        # This must be done BEFORE the FOS driver runs "import uldaq".
        with patch.dict("sys.modules", {"uldaq": Mock()}):
            # Trigger lazy import of the uldaq module.
            qmi.instruments.bristol.fos._import_modules()

        # Patch device descriptor
        self.wrong_dev_desc = Mock()
        self.wrong_dev_desc.unique_id = self.UNIQUE_ID[::-1]
        self.dev_desc = Mock()
        self.dev_desc.unique_id = self.UNIQUE_ID
        patcher = patch('qmi.instruments.bristol.fos.uldaq.get_daq_device_inventory',
                        side_effect=[[self.wrong_dev_desc, self.dev_desc]])
        _ = patcher.start()
        self.addCleanup(patcher.stop)

        # Patch device
        self.device_mock = Mock()
        patcher = patch('qmi.instruments.bristol.fos.uldaq.DaqDevice',
                        return_value=self.device_mock)
        self.device_init = patcher.start()
        self.addCleanup(patcher.stop)

        # Start QMI patcher
        self.qmi_ctx = QMI_Context("test_fos")
        self.qmi_ctx.start("test_fos")

    def tearDown(self):
        # Stop QMI
        self.qmi_ctx.stop()
        logging.getLogger("qmi.instruments.bristol.fos").setLevel(logging.NOTSET)

    def test_open_close(self):
        """Test open & close of the FOS."""
        from qmi.instruments.bristol import Bristol_Fos

        # Pull data types from the mocked uldaq module.
        DigitalPortType = qmi.instruments.bristol.fos.uldaq.DigitalPortType
        DigitalDirection = qmi.instruments.bristol.fos.uldaq.DigitalDirection

        # Arrange
        expected_calls = [
            call.connect(),
            call.get_dio_device(),
            call.disconnect(),
            call.release()
        ]

        expected_port_config = (
            DigitalPortType.FIRSTPORTA, DigitalDirection.OUTPUT
        )

        # Act
        with patch("qmi.instruments.bristol.fos.sys.platform", "linux"):
            fos: Bristol_Fos = Bristol_Fos(self.qmi_ctx, "fosser", self.UNIQUE_ID)
            fos.open()
            fos.close()

        # Assert
        self.assertEqual((self.dev_desc,), self.device_init.call_args[0])
        self.assertEqual(expected_calls, self.device_mock.method_calls)
        self.assertTupleEqual(expected_port_config,
                              self.device_mock.get_dio_device.return_value.d_config_port.call_args[0])

    def test_device_not_found(self):
        """Test exception if device not found."""
        from qmi.instruments.bristol import Bristol_Fos
        # Act
        with patch("qmi.instruments.bristol.fos.sys.platform", "linux"):
            fos: Bristol_Fos = Bristol_Fos(self.qmi_ctx, "fosser", "wrong")

        # Assert
        with self.assertRaises(ValueError):
            fos.open()

    def test_device_connection_errors(self):
        """Test exception if device has connection issues."""
        from qmi.instruments.bristol import Bristol_Fos
        # Arrange
        expected_calls = [
            call.connect(),
            call.get_dio_device(),
            call.disconnect(),
            call.release()
        ]
        self.device_mock.get_dio_device = Mock(return_value=BaseException)
        # Act
        with patch("qmi.instruments.bristol.fos.sys.platform", "linux"):
            fos: Bristol_Fos = Bristol_Fos(self.qmi_ctx, "fosser", self.UNIQUE_ID)
        # Assert
        with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException):
            fos.open()

        self.assertEqual(expected_calls, self.device_mock.method_calls)

    def test_select_channel(self):
        """Test selecting channel"""
        from qmi.instruments.bristol import Bristol_Fos

        # Pull data types from the mocked uldaq module.
        DigitalPortType = qmi.instruments.bristol.fos.uldaq.DigitalPortType

        # Arrange
        test_channel = 1
        expected_d_out = (
            DigitalPortType.FIRSTPORTA, test_channel - 1
        )

        # Act
        with patch("qmi.instruments.bristol.fos.sys.platform", "linux"):
            with qmi.make_instrument("fosser", Bristol_Fos, self.UNIQUE_ID) as fos:
                fos.select_channel(test_channel)

        # Assert
        self.assertTupleEqual(expected_d_out, self.device_mock.get_dio_device.return_value.d_out.call_args[0])

    def test_bad_channel(self):
        """Test bad channel"""
        from qmi.instruments.bristol import Bristol_Fos
        # Act & assert
        with self.assertRaises(ValueError):
            with patch("qmi.instruments.bristol.fos.sys.platform", "linux"):
                with qmi.make_instrument("fosser", Bristol_Fos, self.UNIQUE_ID) as fos:
                    fos.select_channel(0)


@patch("qmi.instruments.bristol.fos.sys.platform", "win32")
class TestFosWindows(unittest.TestCase):
    """ Unit tests for Bristol FOS. """

    UNIQUE_ID = 'DEADBEAF'

    @patch("qmi.instruments.bristol.fos.sys.platform", "win32")
    def setUp(self):
        logging.getLogger("qmi.instruments.bristol.fos").setLevel(logging.CRITICAL)
        # Patch device descriptor
        self.dev_desc = Mock()
        self.dev_desc.unique_id = self.UNIQUE_ID
        self.wrong_dev_desc = Mock()
        self.wrong_dev_desc.unique_id = self.UNIQUE_ID[::-1]
        # Substitute a mock object in place of the "mcculw" module.
        # This must be done BEFORE the FOS driver runs "import mcculw".
        self.mcculw_mock = MagicMock(autospec=mcculw)
        with patch.dict("sys.modules", {
            "mcculw": self.mcculw_mock,
        }):
            # Trigger lazy import of the mcculw module.
            qmi.instruments.bristol.fos._import_modules()

            self.ulpatcher = patch(
                "qmi.instruments.bristol.fos.ul.get_daq_device_inventory",
            )
            self.ulpatcher.start()

        self.addCleanup(self.ulpatcher.stop)

        # Start QMI patcher
        self.qmi_ctx = QMI_Context("test_fos")
        self.qmi_ctx.start("test_fos")

    def tearDown(self):
        # Stop QMI
        self.qmi_ctx.stop()
        logging.getLogger("qmi.instruments.bristol.fos").setLevel(logging.NOTSET)

    def test_open_close(self):
        """Test open & close of the FOS."""
        # Arrange
        from qmi.instruments.bristol import Bristol_Fos
        board_id = 0

        fos: Bristol_Fos = Bristol_Fos(self.qmi_ctx, "fosser", self.UNIQUE_ID)

        self.ulpatcher.target.get_daq_device_inventory.return_value = [self.wrong_dev_desc, self.dev_desc]
        self.ulpatcher.target.d_config_port.side_effect = [None]

        # Act
        with patch("qmi.instruments.bristol.fos.ul.get_board_number", return_value=board_id) as bid_patch:
            fos.open()

        fos.close()

        # Assert
        self.ulpatcher.target.get_daq_device_inventory.assert_called()
        self.ulpatcher.target.create_daq_device.assert_called()
        self.assertEqual(1, bid_patch.call_count)
        self.ulpatcher.target.d_config_port.assert_called()
        self.ulpatcher.target.release_daq_device.assert_called()

    def test_device_not_found(self):
        """Test exception if device not found."""
        # Arrange
        from qmi.instruments.bristol import Bristol_Fos
        # Act
        fos: Bristol_Fos = Bristol_Fos(self.qmi_ctx, "fosser", "wrong")

        # Assert
        with self.assertRaises(ValueError):
            fos.open()

    def test_invalid_board_number(self):
        """Test assertion error is raised, if the board number is not valid."""
        # Arrange
        from qmi.instruments.bristol import Bristol_Fos
        invalid_bids = [-1, 100]
        # Act
        for bid in invalid_bids:
            # Assert
            with self.assertRaises(AssertionError):
                fos = Bristol_Fos(self.qmi_ctx, "fosser", self.UNIQUE_ID, board_id=bid)
                fos._is_open = True
                fos.close()

            fos: Bristol_Fos = Bristol_Fos(self.qmi_ctx, "fosser", self.UNIQUE_ID)
            with self.assertRaises(AssertionError):
                fos.fos.board_id = bid

    def test_device_connection_errors(self):
        """Test exception if device has connection issues."""
        from qmi.instruments.bristol import Bristol_Fos
        # Arrange
        self.ulpatcher.target.get_daq_device_inventory.return_value = [self.dev_desc]
        board_id = 2
        # Act
        fos: Bristol_Fos = Bristol_Fos(self.qmi_ctx, "fosser", self.UNIQUE_ID)
        fos.fos.board_id = board_id

        with patch("qmi.instruments.bristol.fos.ul.get_board_number", return_value=board_id) as bid_patch:
            # Assert
            self.ulpatcher.target.d_config_port.side_effect = [Exception("Fail")]
            with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException):
                fos.open()

        self.ulpatcher.target.create_daq_device.assert_called()
        self.assertEqual(1, bid_patch.call_count)
        self.ulpatcher.target.d_config_port.assert_called()
        self.ulpatcher.target.release_daq_device.assert_called()

    def test_select_channel(self):
        """Test selecting channel."""
        from qmi.instruments.bristol import Bristol_Fos

        # Pull data types from the mocked uldaq module.
        DigitalPortType = qmi.instruments.bristol.fos.enums.DigitalPortType
        # Arrange
        test_channel = 1
        board_id = 0
        self.ulpatcher.target.d_config_port.side_effect = [None]
        self.ulpatcher.target.get_daq_device_inventory.return_value = [self.dev_desc]
        # Act
        with patch("qmi.instruments.bristol.fos.ul.get_board_number", return_value=board_id):
            with qmi.make_instrument("fosser", Bristol_Fos, self.UNIQUE_ID) as fos:
                fos.select_channel(test_channel)

        # Assert
        self.ulpatcher.target.d_out.assert_called_once_with(
            board_id, DigitalPortType.FIRSTPORTA, test_channel - 1
        )

    def test_bad_channel(self):
        """Test bad channel."""
        from qmi.instruments.bristol import Bristol_Fos
        # Act & assert
        board_id = 0
        self.ulpatcher.target.get_daq_device_inventory.return_value = [self.dev_desc]
        with self.assertRaises(ValueError):
            with patch("qmi.instruments.bristol.fos.ul.get_board_number", return_value=board_id):
                with qmi.make_instrument("fosser", Bristol_Fos, self.UNIQUE_ID) as fos:
                    fos.select_channel(0)



"""
Quantum Composers, delay pulse generator.

The qmi.instruments.quantum_composers package provides support for:
- 9530 series.
"""

from qmi.instruments.quantum_composers.pulse_generator9530 import QuantumComposers_PulseGenerator9530 as\
    QuantumComposers_9530
from qmi.instruments.quantum_composers.pulse_generator9530 import RefClkSource, PulseMode, TriggerMode, TriggerEdge,\
    OutputDriver



import random
import unittest
import unittest.mock
import qmi
from qmi.core.exceptions import QMI_InstrumentException, QMI_TimeoutException
from qmi.core.transport import QMI_SerialTransport
from qmi.instruments.thorlabs import Thorlabs_Mpc320
from qmi.instruments.thorlabs.apt_protocol import (
    AptChannelJogDirection,
    AptChannelState,
)
from tests.patcher import PatcherQmiContext


class TestThorlabsMPC320(unittest.TestCase):
    def setUp(self):
        # Patch QMI context and make instrument
        self._ctx_qmi_id = f"test-tasks-{random.randint(0, 100)}"
        self.qmi_patcher = PatcherQmiContext()
        self.qmi_patcher.start(self._ctx_qmi_id)
        self._transport_mock = unittest.mock.MagicMock(spec=QMI_SerialTransport)
        with unittest.mock.patch(
            "qmi.instruments.thorlabs.mpc320.create_transport",
            return_value=self._transport_mock,
        ):
            self._instr: Thorlabs_Mpc320 = self.qmi_patcher.make_instrument(
                "test_mpc320", Thorlabs_Mpc320, "serial:transport_str"
            )
        self._instr.open()

    def tearDown(self):
        self._instr.close()
        qmi.stop()

    def test_get_idn_sends_command_and_returns_identification_info(self):
        """Test get_idn method and returns identification info."""
        # Arrange
        expected_idn = ["Thorlabs", b"MPC320 ", 94000009, 3735810]
        # \x89\x53\x9a\x05 is 94000009
        # \x4d\x50\x43\x33\x32\x30\x0a is MPC320
        # x2c\x00 is Brushless DC controller card
        # \x02\x01\x39\x00 is 3735810
        self._transport_mock.read.side_effect = [
            b"\x06\x00\x54\x00\x00\x81",
            b"\x89\x53\x9a\x05\x4d\x50\x43\x33\x32\x30\x20\x00\x2c\x00\x02\x01\x39\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00",
        ]

        # Act
        idn = self._instr.get_idn()

        # Assert
        self.assertEqual(idn.vendor, expected_idn[0])
        self.assertEqual(idn.model, expected_idn[1])
        self.assertEqual(idn.serial, expected_idn[2])
        self.assertEqual(idn.version, expected_idn[3])

        self._transport_mock.write.assert_called_once_with(bytearray(b"\x05\x00\x00\x00P\x01"))
        self._transport_mock.read.assert_has_calls(
            [
                unittest.mock.call(nbytes=6, timeout=1.0),
                unittest.mock.call(nbytes=84, timeout=1.0),
            ]
        )

    def test_get_idn_with_wrong_returned_msg_id_sends_command_and_throws_error(self):
        """Test get_idn method and returns identification info."""
        # Arrange
        self._transport_mock.read.side_effect = [
            b"\x11\x00\x54\x00\x00\x81",
            b"\x89\x53\x9a\x05\x4d\x50\x43\x33\x32\x30\x20\x00\x2c\x00\x02\x01\x39\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00",
        ]

        # Act
        # Assert
        with self.assertRaises(QMI_InstrumentException):
            _ = self._instr.get_idn()
        self._transport_mock.write.assert_called_once_with(bytearray(b"\x05\x00\x00\x00P\x01"))
        self._transport_mock.read.assert_has_calls(
            [
                unittest.mock.call(nbytes=6, timeout=1.0),
                unittest.mock.call(nbytes=84, timeout=1.0),
            ]
        )

    def test_identify_sends_command(self):
        """Test identify method and send relevant command."""
        # Arrange

        # Act
        self._instr.identify()

        # Assert
        self._transport_mock.write.assert_called_once_with(bytearray(b"\x23\x02\x01\x00\x50\x01"))

    def test_enable_channels_enable_four_throws_exception(self):
        """Test enable channel 4, throws exception."""
        # Arrange

        # Act
        # Assert
        with self.assertRaises(QMI_InstrumentException):
            self._instr.enable_channels([4])

    def test_enable_channels_enable_one_sends_command(self):
        """Test enable channel 1, send command to enable channel 1."""
        # Arrange

        # Act
        self._instr.enable_channels([1])

        # Assert
        self._transport_mock.write.assert_called_once_with(bytearray(b"\x10\x02\x01\x01\x50\x01"))

    def test_enable_channels_enable_one_and_three_sends_command(self):
        """Test enable channels 1 and 3, sends command to enable channels 1 and 3."""
        # Arrange

        # Act
        self._instr.enable_channels([1, 3])

        # Assert
        self._transport_mock.write.assert_called_once_with(bytearray(b"\x10\x02\x05\x01\x50\x01"))

    def test_disable_all_channels_sends_command(self):
        """Test disable all channels, sends command to disable all channels."""
        # Arrange

        # Act
        self._instr.disable_all_channels()

        # Assert
        self._transport_mock.write.assert_called_once_with(bytearray(b"\x10\x02\x00\x01\x50\x01"))

    def test_get_channel_state_for_channel_four_throws_exception(self):
        """Test get state of channel 4, throws exception."""
        # Arrange

        # Act
        # Assert
        with self.assertRaises(QMI_InstrumentException):
            self._instr.get_channel_state(4)

    def test_get_channel_state_for_disabled_channel_1_sends_command_and_returns_disable_value(
        self,
    ):
        """Test get state of disabled channel 1, sends command to get state."""
        # Arrange
        self._transport_mock.read.return_value = b"\x12\x02\x01\x00\x00\x81"

        # Act
        state = self._instr.get_channel_state(1)

        # Assert
        self.assertEqual(state, AptChannelState.DISABLE)
        self._transport_mock.write.assert_called_once_with(bytearray(b"\x11\x02\x01\x00\x50\x01"))

    def test_get_channel_state_for_enabled_channel_1_sends_command_and_returns_enabled_value(
        self,
    ):
        """Test get state of enabled channel 1, sends command to get state."""
        # Arrange
        self._transport_mock.read.return_value = b"\x12\x02\x01\x01\x00\x81"

        # Act
        state = self._instr.get_channel_state(1)

        # Assert
        self.assertEqual(state, AptChannelState.ENABLE)
        self._transport_mock.write.assert_called_once_with(bytearray(b"\x11\x02\x01\x00\x50\x01"))

    def test_get_channel_state_for_disabled_channel_2_sends_command_and_returns_disabled_value(
        self,
    ):
        """Test get state of disabled channel 2, sends command to get state.
        This test checks that if anything but 0x01 is received then the channel is disabled."""
        # Arrange
        self._transport_mock.read.return_value = b"\x12\x02\x01\x03\x00\x81"

        # Act
        state = self._instr.get_channel_state(1)

        # Assert
        self.assertEqual(state, AptChannelState.DISABLE)
        self._transport_mock.write.assert_called_once_with(bytearray(b"\x11\x02\x01\x00\x50\x01"))

    def test_start_auto_status_update_sends_command(self):
        """Test start automatic status updates, sends command"""
        # Arrange

        # Act
        self._instr.start_auto_status_update()

        # Assert
        self._transport_mock.write.assert_called_once_with(bytearray(b"\x11\x00\x00\x00\x50\x01"))

    def test_stop_auto_status_update_sends_command(self):
        """Test stop automatic status updates, sends command"""
        # Arrange

        # Act
        self._instr.stop_auto_status_update()

        # Assert
        self._transport_mock.write.assert_called_once_with(bytearray(b"\x12\x00\x00\x00\x50\x01"))

    def test_home_channel_3_sends_command(self):
        """Test home channel 3, sends command"""
        # Arrange

        # Act
        self._instr.home_channel(3)

        # Assert
        self._transport_mock.write.assert_called_once_with(bytearray(b"\x43\x04\x04\x00\x50\x01"))

    def test_is_channel_1_homed_when_channel_homed_sends_command_returns_status(self):
        """Test is_channel_homed for channel 1 when channel is homed, send command and return homed value."""
        # Arrange
        # first 2 binary strings are responses from the status update command
        # last string is the homed response
        self._transport_mock.read.side_effect = [
            b"\x91\x04\x0e\x00\x00\x81",
            b"\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00",
            b"\x44\x04\x01\x00\x00\x81",
        ]

        # Act
        state = self._instr.is_channel_homed(1)

        # Assert
        self.assertTrue(state)
        self._transport_mock.write.assert_not_called()
        self._transport_mock.read.assert_has_calls(
            [
                unittest.mock.call(nbytes=6, timeout=1.0),
                unittest.mock.call(nbytes=14, timeout=1.0),
                unittest.mock.call(nbytes=6, timeout=1.0),
            ]
        )

    def test_is_channel_1_homed_with_read_timing_out_sends_command_returns_status(self):
        """Test is_channel_homed for channel 1 when read times out, send command and returns not homed."""
        # Arrange
        # first 2 binary strings are responses from the status update command
        # last string is the homed response
        self._transport_mock.read.side_effect = [
            b"\x91\x04\x0e\x00\x00\x81",
            b"\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00",
            QMI_TimeoutException(),
        ]

        # Act
        state = self._instr.is_channel_homed(1)

        # Assert
        self.assertFalse(state)
        self._transport_mock.write.assert_not_called()
        self._transport_mock.read.assert_has_calls(
            [
                unittest.mock.call(nbytes=6, timeout=1.0),
                unittest.mock.call(nbytes=14, timeout=1.0),
                unittest.mock.call(nbytes=6, timeout=1.0),
            ]
        )

    def test_move_absolute_sends_move_command(self):
        """Test move channel 1, sends move command."""
        # Arrange
        expected_position = 10

        # Act
        self._instr.move_absolute(1, expected_position)

        # Assert
        self._transport_mock.write.assert_called_once_with(
            bytearray(b"\x53\x04\x06\x00\xd0\x01\x01\x00\x51\x00\x00\x00")
        )

    def test_move_absolute_outside_range_thorws_error(self):
        """Test move channel 1 outside valid range, throws error."""
        # Arrange

        # Act
        # Assert
        with self.assertRaises(QMI_InstrumentException):
            self._instr.move_absolute(1, 20000)

    def test_is_channel_1_move_completed_when_channel_move_completed_sends_command_returns_status(
        self,
    ):
        """Test is_move_completed for channel 1 when channel is moved, send command and return move completed value."""
        # Arrange
        # first 2 binary strings are responses from the status update command
        # last string is the homed response
        self._transport_mock.read.side_effect = [
            b"\x91\x04\x0e\x00\x00\x81",
            b"\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00",
            b"\x64\x04\x01\x00\x00\x81",
        ]

        # Act
        state = self._instr.is_move_completed(1)

        # Assert
        self.assertTrue(state)
        self._transport_mock.write.assert_not_called()
        self._transport_mock.read.assert_has_calls(
            [
                unittest.mock.call(nbytes=6, timeout=1.0),
                unittest.mock.call(nbytes=14, timeout=1.0),
                unittest.mock.call(nbytes=6, timeout=1.0),
            ]
        )

    def test_is_channel_1_move_completed_with_read_timing_out_sends_command_returns_status(
        self,
    ):
        """Test is_move_completed for channel 1 when read times out, sends command and returns move not completed."""
        # Arrange
        # first 2 binary strings are responses from the status update command
        # last string is the homed response
        self._transport_mock.read.side_effect = [
            b"\x91\x04\x0e\x00\x00\x81",
            b"\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00",
            QMI_TimeoutException,
        ]

        # Act
        state = self._instr.is_move_completed(1)

        # Assert
        self.assertFalse(state)
        self._transport_mock.write.assert_not_called()
        self._transport_mock.read.assert_has_calls(
            [
                unittest.mock.call(nbytes=6, timeout=1.0),
                unittest.mock.call(nbytes=14, timeout=1.0),
                unittest.mock.call(nbytes=6, timeout=1.0),
            ]
        )

    def test_save_parameter_settings_for_given_command_sends_command(self):
        """Test save_parameter_settings for channel 1 for a specific command, send command."""
        # Arrange

        # Act
        self._instr.save_parameter_settings(1, 0x04B6)

        # Assert
        self._transport_mock.write.assert_called_once_with(bytearray(b"\xb9\x04\x04\x00\xd0\x01\x01\x00\xb6\x04"))

    def test_get_status_updated_of_channel_1_sends_command_returns_status(
        self,
    ):
        """Test get_status_update for channel 1, send command and return status."""
        # Arrange
        self._transport_mock.read.side_effect = [
            b"\x91\x04\x0e\x00\x00\x81",
            b"\x01\x00\x51\x00\x00\x00\x00\x00\xff\xff\x00\x00\x00\x00",
        ]

        # Act
        status = self._instr.get_status_update(1)

        # Assert
        self.assertEqual(status.channel, 1)
        self.assertEqual(round(status.position), 10)
        self.assertEqual(status.velocity, 0)
        self.assertEqual(status.motor_current, -1)
        self._transport_mock.write.assert_called_once_with(bytearray(b"\x90\x04\x01\x00\x50\x01"))
        self._transport_mock.read.assert_has_calls(
            [
                unittest.mock.call(nbytes=6, timeout=1.0),
                unittest.mock.call(nbytes=14, timeout=1.0),
            ]
        )

    def test_jog_forward_sends_command(self):
        """Test jog forward for channel 1, sends command."""
        # Arrange

        # Act
        self._instr.jog(1, AptChannelJogDirection.FORWARD)

        # Assert
        self._transport_mock.write.assert_called_once_with(bytearray(b"\x6a\x04\x01\x01\x50\x01"))

    def test_jog_backward_sends_command(self):
        """Test jog backward for channel 1, sends command."""
        # Arrange

        # Act
        self._instr.jog(1, AptChannelJogDirection.BACKWARD)

        # Assert
        self._transport_mock.write.assert_called_once_with(bytearray(b"\x6a\x04\x01\x02\x50\x01"))

    def test_set_polarisation_parameters_sends_command(self):
        """Test set_polarisation_parameters, sends command."""
        # Arrange
        vel = 50
        home_pos = 85
        jog_step1 = 0
        jog_step2 = 17
        jog_step3 = 34

        # Act
        self._instr.set_polarisation_parameters(
            vel,
            home_pos,
            jog_step1,
            jog_step2,
            jog_step3,
        )

        # Assert
        self._transport_mock.write.assert_called_once_with(
            bytearray(b"\x30\x05\x0c\x00\xd0\x01\x00\x00\x32\x00\xad\x02\x00\x00\x89\x00\x12\x01")
        )

    def test_set_polarisation_parameters_with_invalid_velocity_throws_error(self):
        """Test set_polarisation_parameters with invalid velocity, throws error."""
        # Arrange
        vel = 120
        home_pos = 85
        jog_step1 = 0
        jog_step2 = 17
        jog_step3 = 34

        # Act
        # Assert
        with self.assertRaises(QMI_InstrumentException):
            self._instr.set_polarisation_parameters(
                vel,
                home_pos,
                jog_step1,
                jog_step2,
                jog_step3,
            )

    def test_set_polarisation_parameters_with_invalid_home_position_throws_error(self):
        """Test set_polarisation_parameters with invalid home position, throws error."""
        # Arrange
        vel = 50
        home_pos = 200
        jog_step1 = 0
        jog_step2 = 17
        jog_step3 = 34

        # Act
        # Assert
        with self.assertRaises(QMI_InstrumentException):
            self._instr.set_polarisation_parameters(
                vel,
                home_pos,
                jog_step1,
                jog_step2,
                jog_step3,
            )

    def test_set_polarisation_parameters_with_invalid_jog_step_1_throws_error(self):
        """Test set_polarisation_parameters with invalid jog step 1, throws error."""
        # Arrange
        vel = 50
        home_pos = 85
        jog_step1 = 300
        jog_step2 = 17
        jog_step3 = 34

        # Act
        # Assert
        with self.assertRaises(QMI_InstrumentException):
            self._instr.set_polarisation_parameters(
                vel,
                home_pos,
                jog_step1,
                jog_step2,
                jog_step3,
            )

    def test_set_polarisation_parameters_with_invalid_jog_step_2_throws_error(self):
        """Test set_polarisation_parameters with invalid jog step 2, throws error."""
        # Arrange
        vel = 50
        home_pos = 85
        jog_step1 = 0
        jog_step2 = 240
        jog_step3 = 34

        # Act
        # Assert
        with self.assertRaises(QMI_InstrumentException):
            self._instr.set_polarisation_parameters(
                vel,
                home_pos,
                jog_step1,
                jog_step2,
                jog_step3,
            )

    def test_set_polarisation_parameters_with_invalid_jog_step_3_throws_error(self):
        """Test set_polarisation_parameters with invalid jog step 3, throws error."""
        # Arrange
        vel = 50
        home_pos = 85
        jog_step1 = 0
        jog_step2 = 17
        jog_step3 = 310

        # Act
        # Assert
        with self.assertRaises(QMI_InstrumentException):
            self._instr.set_polarisation_parameters(
                vel,
                home_pos,
                jog_step1,
                jog_step2,
                jog_step3,
            )

    def test_get_polarisation_parameters_sends_command_returns_parameters(
        self,
    ):
        """Test get_polarisation_parameters, sends command and returns parameters."""
        # Arrange
        expected_velocity = 50
        expected_home_position = 85
        expected_jog_step_1 = 0
        expected_jog_step_2 = 17
        expected_jog_step_3 = 34
        self._transport_mock.read.side_effect = [
            b"\x32\x05\x0c\x00\x00\x81",
            b"\x00\x00\x32\x00\xad\x02\x00\x00\x89\x00\x12\x01",
        ]

        # Act
        params = self._instr.get_polarisation_parameters()

        # Assert
        self.assertEqual(params.velocity, expected_velocity)
        self.assertEqual(params.home_position, expected_home_position)
        self.assertEqual(params.jog_step1, expected_jog_step_1)
        self.assertEqual(params.jog_step2, expected_jog_step_2)
        self.assertEqual(params.jog_step3, expected_jog_step_3)
        self._transport_mock.write.assert_called_once_with(bytearray(b"\x31\x05\x00\x00\x50\x01"))
        self._transport_mock.read.assert_has_calls(
            [
                unittest.mock.call(nbytes=6, timeout=1.0),
                unittest.mock.call(nbytes=12, timeout=1.0),
            ]
        )


if __name__ == "__main__":
    unittest.main()



""" QMI instrument driver for the High Finesse Wavemeter instrument.

The instrument driver makes use of the manufacturer provided software libraries, "libwlmData.so" for Linux OS,
or "wlmData.dll" for Windows, or "libwlmData.dylib" for MacOS.
Please find the licence terms for these files as well as further documentation on the manufacturer's website at
https://www.highfinesse.com/en/support/downloads.html.
This driver has been tested on the WS-6 model.
"""

import ctypes
import logging

import numpy as np

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method
from qmi.instruments.high_finesse.support import wlmConst
from qmi.instruments.high_finesse.support._library_wrapper import _LibWrapper, WlmGetErr

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class HighFinesse_Wlm(QMI_Instrument):
    """A network based driver for the High Finesse Wavelength meter.

    This driver automatically detects the platform of the client and will load the respective driver library
    (wlmData.dll for Windows, libwlmData.so for Linux, and libwlmData.dylib for MacOS). Make sure that the library is
    available on your system and that the wlmData.ini file contains the IP address of the server.
    """
    # TODO: This could set with GetChannelsCount call after super().open().
    MAX_CHANNEL_NUMBER = 8

    def __init__(self, context: QMI_Context, name: str) -> None:
        """Initialize the instrument driver.

        Parameters:
            context: QMI_Context object for the instrument driver.
            name:    Name for this instrument instance.
        """
        self._lib: _LibWrapper = _LibWrapper()
        super().__init__(context, name)

    def _check_channel(self, channel: int) -> None:
        """A method for checking the channel number.

        Raises:
            QMI_InstrumentException: In case of instrument error or channel out of range.
        """
        self._check_is_open()
        if not (0 < channel <= self.MAX_CHANNEL_NUMBER):
            raise QMI_InstrumentException(f"Channel number out of range: {channel}")

    def _check_for_error_code(self, value: float | int, method: str) -> float | int:
        """A method for checking error codes returned by the device.

        Parameters:
            value:  Value to check for an error code.
            method: The called method to check the value for.

        Returns:
            value:  If value is larger than 0, or NaN, it cannot be an error code but a result value.

        Raises:
            QMI_InstrumentException: If the value to check is an error code, raise.
        """
        # TODO: Expand in future also for other error types, which are call dependent. Now only WlmGetErr.
        if value > 0 or value is np.nan:
            return value

        i_val = int(value)
        err_msg = WlmGetErr(i_val) if i_val in [error.value for error in WlmGetErr] else str(value)
        _logger.error("[%s] %s error: %s", self._name, method, err_msg)

        raise QMI_InstrumentException(f"Error received from library call '{method}': {err_msg}")

    @rpc_method
    def open(self) -> None:
        _logger.debug("[%s] Opening connection to instrument", self._name)
        super().open()

    @rpc_method
    def close(self) -> None:
        _logger.info("[%s] Closing connection to instrument", self._name)
        super().close()

    @rpc_method
    def start_server(self) -> bool:
        """Start the device server if it is not running already.

        Returns:
             True if the server was started, or False if already running.
        """
        wlm_count = self._lib.dll.GetWLMCount(0)
        _logger.info("[%s] Starting server, current wlm count = %d", self._name, wlm_count)
        if wlm_count == 0:
            status = self._lib.dll.ControlWLMEx(wlmConst.cCtrlWLMWait, 0, 0, 10_000, 0)
            _logger.debug("[%s] Started server with status = %d", self._name, status)
            status = self._lib.dll.Operation(wlmConst.cCtrlStartMeasurement)
            _logger.debug("[%s] Started measurement with status = %d", self._name, status)
            return True

        return False

    @rpc_method
    def stop_server(self) -> None:
        """Stop the device server if it is running."""
        wlm_count = self._lib.dll.GetWLMCount(0)
        _logger.info("[%s] Stopping server, current wlm count = %d", self._name, wlm_count)
        if wlm_count > 0:
            status = self._lib.dll.Operation(wlmConst.cCtrlStopAll)
            _logger.debug("[%s] Stopped measurement with status = %d", self._name, status)
            status = self._lib.dll.ControlWLM(wlmConst.cCtrlWLMExit, 0, 0)
            _logger.debug("[%s] Stopped server with status = %d", self._name, status)

    @rpc_method
    def get_version(self) -> str:
        """Get the WLM version as a string.

        Returns:
            str: The version in the format "WLM Version: [{type}.{version}.{revision}.{build}]"
        """
        self._check_is_open()
        version_type = self._lib.dll.GetWLMVersion(0)
        version_ver = self._lib.dll.GetWLMVersion(1)
        version_rev = self._lib.dll.GetWLMVersion(2)
        version_build = self._lib.dll.GetWLMVersion(3)

        version_str = f"{version_type}.{version_ver}.{version_rev}.{version_build}"
        _logger.info("[%s] Get version: [%s]", self._name, version_str)

        return f"WLM Version: [{version_str}]"

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """Read instrument type and version and return QMI_InstrumentIdentification instance."""
        version_str = self.get_version()
        model, version, revision = version_str[14:].split(".")[:3]
        return QMI_InstrumentIdentification(
            vendor="HighFinesse",
            model=f"WLM-{model}",
            serial=None,
            version=".".join([version, revision])
        )

    @rpc_method
    def get_operation_state(self) -> int:
        """Get the instrument operation state. Possible return values are:
        - 0 : cStop. Wlm active but stopped,
        - 1 : cAdjustment. Wlm active and program is adjusting,
        - 2 : cMeasurement. Wlm active and measuring, recording or replaying.

        Returns:
            op_state: The operation state integer.
        """
        self._check_is_open()
        op_state = self._lib.dll.GetOperationState(0)
        return op_state

    @rpc_method
    def get_frequency(self, channel: int) -> float:
        """Get the main results of the measurement of a specified signal.

        Parameters:
            channel: The signal number (1 to 8) in case of a WLM with multichannel switch or with double pulse
                     option (MLC). For WLMs without these options 1 should be overhanded.

        Returns:
            frequency: The last measured frequency value in THz.

        Raises:
            QMI_InstrumentException: In case of instrument error.
        """
        _logger.debug("[%s] Getting frequency on channel %d", self._name, channel)
        self._check_channel(channel)

        frequency = self._lib.dll.GetFrequencyNum(channel, 0.0)

        return self._check_for_error_code(frequency, "GetFrequencyNum")

    @rpc_method
    def get_wavelength(self, channel: int) -> float:
        """Get the main results of the measurement of a specified signal.

        Parameters:
            channel: The signal number (1 to 8) in case of a WLM with multichannel switch or with double pulse
                     option (MLC). For WLMs without these options 1 should be overhanded.

        Returns:
            wavelength: The last measured wavelength in nm.
        """
        _logger.debug("[%s] Getting wavelength on channel %d", self._name, channel)
        self._check_channel(channel)

        wavelength = self._lib.dll.GetWavelengthNum(channel, 0.0)

        return self._check_for_error_code(wavelength, "GetWavelengthNum")

    @rpc_method
    def get_power(self, channel: int) -> float:
        """Get the power of the current measurement shot of a specified signal.

        Parameters:
            channel: The signal number (1 to 8) in case of a WLM with multichannel switch or with double pulse
                     option (MLC). For WLMs without these options 1 should be overhanded.

        Returns:
            power:   The power of the last measured cw or quasi cw signal in uW or the energy in uJ.
        """
        _logger.debug("[%s] Getting power on channel %d", self._name, channel)
        self._check_channel(channel)

        power = self._lib.dll.GetPowerNum(channel, 0.0)

        return self._check_for_error_code(power, "GetPowerNum")

    @rpc_method
    def set_data_pattern(self, index: int = 1, enable: bool = True) -> None:
        """Enable or disable a pattern index to export data from instrument.

        Parameters:
            index:  Index number for export control. Index values have the following meanings:
                    - 0 : cSignal1Interferometers. The array received by the Fizeau interferometers or
                      diffraction grating.
                    - 1 : cSignal1WideInterferometer. Additional long interferometer or grating array.
                    - 1 : cSignal1Grating. Only in Grating analyzing versions! The array received
                      by spectrum analysis (grating precision).
                    - 2 : cSignal2Interferometers. Only in Double Pulse versions! The array received by
                      the Fizeau interferometers for the 2nd pulse.
                    - 3 : cSignal2WideInterferometer. Only in Double Pulse versions! Additional long
                      interferometer array for 2nd pulse.

                    With SetAnalysis only the following value is possible (with LSA and resolver versions):
                    - 4 : cSignalAnalysis[X]. The [X] array[s] of the spectral analysis data.
                    - 5 : cSignalAnalysisY. The Y array of the spectral analysis data.

            enable: True to enable data pattern (default), False to disable.

        Raises:
            ValueError: With invalid index number.
        """
        if index not in range(6):
            raise ValueError(f"Invalid data pattern index number {index}")

        if enable:
            _logger.debug("[%s] Enabling data pattern with index %d", self._name, index)
            i_enable = wlmConst.cPatternEnable

        else:
            _logger.debug("[%s] Disabling data pattern with index %d", self._name, index)
            i_enable = wlmConst.cPatternDisable

        ret_val = self._lib.dll.SetPattern(index, i_enable)
        self._check_for_error_code(ret_val, "SetPattern")

    @rpc_method
    def get_data_pattern_data(
        self, channel: int, index: int, arr_size: int
    ) -> np.typing.NDArray[np.int16]:
        """Get data from the given data pattern. It must be set first with `set_data_pattern`.

        Parameters:
            channel:  The signal number (1 to 8) in case of a WLM with multichannel switch or with double pulse
                      option (MLC). For WLMs without these options 1 should be overhanded.
            index:    Index number to get the data pattern from. Must be in range [0 .. 5]. See also 'set_data_pattern'.
            arr_size: A batch size for reading data pattern.

        Returns:
            pattern_array: Data array from the data pattern.

        Raises:
            ValueError: With invalid index number.
        """
        _logger.debug("[%s] Getting data pattern data on channel %d and index %i", self._name, channel, index)
        if index not in range(6):
            raise ValueError(f"Invalid data pattern index number {index}")

        self._check_channel(channel)
        pattern_array = np.zeros(arr_size, dtype=ctypes.c_short)
        c_short_ptr = ctypes.POINTER(ctypes.c_short)
        ret_val = self._lib.dll.GetPatternDataNum(channel, index, pattern_array.ctypes.data_as(c_short_ptr))
        if ret_val == 0:
            # The channel and/or index was not (correctly) enabled.
            _logger.debug("[%s] GetPatternDataNum error, perhaps SetPattern was not run first?", self._name)

        elif ret_val not in [0, 1]:
            # Return value is an error code.
            self._check_for_error_code(ret_val, "GetPatternDataNum")

        return pattern_array



"""Instrument driver for the Parallax USB Propeller, for servo control."""

import logging
import struct
from time import sleep
from typing import List, Optional, Union

from qmi.core.context import QMI_Context
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method
from qmi.core.transport import create_transport

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class Parallax_UsbPropeller(QMI_Instrument):
    """Instrument driver for the Parallax USB Propeller for servo control."""
    DEFAULT_BAUDRATE = 2400
    RESPONSE_TIMEOUT = 1.0
    AVAILABLE_CHANNELS = 16
    MAX_SPEED = 64
    MAX_PULSE_WIDTH = 1024  # Would be about 2.048 ms
    DEFAULT_MIN = 500  # 1 ms
    DEFAULT_MAX = 688  # 1.376 ms

    def __init__(self, context: QMI_Context, name: str, transport: str) -> None:
        super().__init__(context, name)
        self._transport = create_transport(
            transport, default_attributes={"baudrate": self.DEFAULT_BAUDRATE, "timeout": self.RESPONSE_TIMEOUT}
        )
        self._endline = b"\r"
        self._ramp_speed: List[int] = [0] * self.AVAILABLE_CHANNELS  # Keep track of the ramp speed
        # Track target value for each servo. Targets start at 0 (= no move).
        self.targets: List[int] = [0] * self.AVAILABLE_CHANNELS
        # Servo minimum and maximum targets can be restricted to protect components.
        self.minimums: List[int] = [self.DEFAULT_MIN] * self.AVAILABLE_CHANNELS
        self.maximums: List[int] = [self.DEFAULT_MAX] * self.AVAILABLE_CHANNELS

    def _check_channel(self, channel: int) -> None:
        """Check the channel number."""
        if not 0 <= channel < self.AVAILABLE_CHANNELS:
            raise ValueError(f"Invalid channel number. Must be in range 0-{self.AVAILABLE_CHANNELS - 1}")

    def _write(self, cmd: str, channel: Optional[int] = None) -> None:
        """Writes a command to servo. The command consists of 6 character bytes (c), followed by an
        unsigned integer (B) and closed off with a character byte, which is carriage return.

        Parameters:
            cmd:        The string to write.
            channel:    Channel number.
        """
        # Determine struct.pack string
        pack_string = "ccc"  # Start with the obligatory "!SC" string
        bytes_list: List[Optional[bytes]] = [b"!", b"S", b"C"] + [None] * 5
        count = 3
        for c in cmd:
            pack_string += "c"  # add all command string characters
            bytes_list[count] = c.encode()
            count += 1

        if channel is not None:
            pack_string += "B"  # Add channel number integer
            bytes_list[count] = channel  # type: ignore
            count += 1

        pack_string += "c"  # Add endline
        bytes_list[-1] = self._endline
        l = [pack_string] + bytes_list
        message = struct.pack(*l)  # type: ignore
        self._transport.write(message)
        sleep(len(bytes_list) / self.DEFAULT_BAUDRATE)  # Sleep some bits to let the message arrive

    def _write_value(self, channel: int, value: int, default: bool = False) -> None:
        """Set default or new target value. Default value consists of four character bytes (c),
        followed by three unsigned integers (B), closing off with a character byte, which is carriage return.
        Setting new value consists of three character bytes (c), followed by four unsigned integers (B),
        closing off with a character byte, which is carriage return.

        Parameters:
            channel:        Channel number.
            value:          The target value. The modulo 256 of this is the lowbyte and 8-bit right shift the highbyte.
            default:        If setting default value, True. If setting new target value, False.
        """
        target = value
        ramp_speed = self._ramp_speed[channel]
        if target < self.minimums[channel]:
            target = self.minimums[channel]
            _logger.warning("Target value for channel %i below minimum. Setting to %i", channel, target)
        elif self.maximums[channel] < target:
            target = self.maximums[channel]
            _logger.warning("Target value for channel %i above maximum. Setting to %i", channel, target)

        if default:
            message = struct.pack(
                "ccccBBBc", b"!", b"S", b"C", b"D", channel, target % 256, target >> 8, self._endline
            )
        else:
            message = struct.pack(
                "cccBBBBc", b"!", b"S", b"C", channel, ramp_speed, target % 256, target >> 8, self._endline
            )

        self._transport.write(message)
        sleep(8 / self.DEFAULT_BAUDRATE)  # Sleep some bits to let the message arrive

    def _ask(self, cmd: str, channel: Optional[int] = None) -> Union[int, str]:
        """Writing a query and receiving a response. The query consists of 6 character bytes (c), followed by an
        unsigned integer (B) and closed off with a character byte, which is carriage return. This query always
        returns 3 bytes.

        Parameters:
            cmd:        The query string to write.
            channel:    Channel number.

        Returns:
            The position value of the channel left-shifted by 8 bits (highbyte) and added to it is the lowbyte.
        """
        self._transport.discard_read()
        self._write(cmd, channel)
        # Read answer bytes
        result = self._transport.read(nbytes=3, timeout=self.RESPONSE_TIMEOUT)
        read_channel, highbyte, lowbyte = struct.unpack("BBB", result)
        if chr(highbyte) == "." and 48 <= read_channel < 58 and 48 <= lowbyte < 58:
            return chr(read_channel) + chr(highbyte) + chr(lowbyte)  # It is a reply to FW version number request.

        elif chr(read_channel) == "C" and chr(highbyte) == "L" and chr(lowbyte) == "R":
            return "CLR"  # It is a reply to clear channel EEPROM command.

        elif chr(read_channel) + chr(highbyte) in ["BR", "PM", "DL"]:
            return lowbyte  # It is a reply to Baud rate set, port range set or startup servo mode set command

        assert read_channel == channel

        return (highbyte << 8) + lowbyte

    @rpc_method
    def open(self) -> None:
        self._check_is_closed()
        _logger.info("Opening connection to %s", self._name)
        self._transport.open()
        super().open()

    @rpc_method
    def close(self) -> None:
        self._check_is_open()
        _logger.info("Closing connection to %s", self._name)
        super().close()
        self._transport.close()

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """Read instrument type and version and return QMI_InstrumentIdentification instance."""
        self._check_is_open()
        version = str(self._ask("VER?"))
        return QMI_InstrumentIdentification(vendor="Parallax", model="USB Propeller", serial=None, version=version)

    @rpc_method
    def set_target_value(self, channel: int, value: int) -> None:
        """Set Servo channel to move given value time with certain ramp speed. The time value represents a pulse width
        with 2us increments. For example, to send a 1 ms pulse to a channel you would use a pulse width value of 500
        (500 x 2 us = 1000 us = 1 ms).

        Parameters:
            channel:        The channel number.
            value:          The servo target move value [1-1024], in 2us increments.
        """
        self._check_is_open()
        self._check_channel(channel)
        if not 0 < value <= self.MAX_PULSE_WIDTH:
            raise ValueError(f"Invalid move value {value}. Must be in range 1-{self.MAX_PULSE_WIDTH}.")

        self._write_value(channel, value)
        self.targets[channel] = value

    @rpc_method
    def set_default_move(self, channel: int, value: int) -> None:
        """Set Servo channel default move time at start-up. This value is then stored in EEPROM. The time value
        represents a pulse width with 2us increments. For example, to send a 1 ms pulse to a channel you would use a
        pulse width value of 500 (500 x 2 us = 1000 us = 1 ms).

        Parameters:
            channel:        The channel number.
            value:          The servo target move value [1-1024].
        """
        self._check_is_open()
        self._check_channel(channel)
        if not 0 < value <= self.MAX_PULSE_WIDTH:
            raise ValueError(f"Invalid move value {value}. Must be in range 1-{self.MAX_PULSE_WIDTH}.")

        self._write_value(channel, value, default=True)

    @rpc_method
    def get_target_values(self) -> List[int]:
        """Return the current target values list."""
        return self.targets

    @rpc_method
    def get_value(self, channel: int) -> int:
        """Return current value of a servo channel.

        Parameters:
            channel:        The channel number.

        Returns:
            The channel position.
        """
        self._check_is_open()
        self._check_channel(channel)
        return int(self._ask("RSP", channel))

    @rpc_method
    def set_speed(self, channel: int, speed: int) -> None:
        """Set speed of channel in steps between 0 and 64. This option specifies the speed of the servo.

        Parameters:
            channel:    The channel number.
            speed:      The servo ramp speed value [0-63].


        Raises:
            ValueError if speed not in range 0-1000.
        """
        self._check_is_open()
        self._check_channel(channel)
        if not 0 <= speed < self.MAX_SPEED:
            raise ValueError(f"Invalid ramp speed value {speed}. Must be in range 0-{self.MAX_SPEED -1}")

        self._ramp_speed[channel] = speed

    @rpc_method
    def get_speed(self, channel: int) -> int:
        """Get ramp speed of a channel.

        Parameters:
            channel:    The channel number.

        Returns:
            speed: The servo ramp speed value of the channel [0-63].
        """
        return self._ramp_speed[channel]

    @rpc_method
    def set_baud_rate(self, baud_rate: int) -> None:
        """Set the servo baud rate.

        Parameters:
            baud_rate:      The new baud rate, either as 0, 1, 2400 or 38400. 0 == 2400 and 1 == 38400

        Raises:
            ValueError if input baud_rate is not valid.
            AssertionError if returned baud_rate_set is not the same as the set baud rate.
        """
        self._check_is_open()
        valid_values = [0, 1, 2400, 38400]
        if baud_rate not in valid_values:
            raise ValueError(f"Invalid baud rate {baud_rate}. Should be one of valid values {valid_values}")

        # As we can give as input 0 or 1 only, set to 0 or 1
        if baud_rate == 2400:
            baud_rate = 0
        elif baud_rate == 38400:
            baud_rate = 1

        baud_rate_set = self._ask("SBR", baud_rate)
        if baud_rate == 0:
            assert baud_rate_set == 0

        else:
            assert baud_rate_set == 1

    @rpc_method
    def set_software_port_range(self, port_range: int) -> None:
        """Set software port range to 0-15 or 16-31.

        Parameters:
            port_range:     The new port range, either as 0 for 0-15 or 1 for 16-31.

        Raises:
            ValueError if input port_range is not valid.
            AssertionError if returned port_range_set is not the same as the set port range.
        """
        self._check_is_open()
        if port_range not in [0, 1]:
            raise ValueError(f"Invalid port range {port_range}. The valid values are 0 for 0-15 or 1 for 16-31.")

        port_range_set = self._ask("PSS", port_range)
        assert port_range_set == port_range

    @rpc_method
    def disable_servo_channel(self, channel: int) -> None:
        """Disable a servo channel. This will cause any connected servo to become lax and not try to hold its position.

        Parameters:
            channel:        The channel number.
        """
        self._check_is_open()
        self._check_channel(channel)
        self._write("PSD", channel)

    @rpc_method
    def enable_servo_channel(self, channel: int) -> None:
        """Enable a servo channel. Enabling a channel will cause it to move to the last position it was commanded to,
        or the startup default if no other position commands have been sent since power up/reset.

        Parameters:
            channel:        The channel number.
        """
        self._check_is_open()
        self._check_channel(channel)
        self._write("PSE", channel)

    @rpc_method
    def set_startup_servo_mode(self, servo_mode: int) -> None:
        """Set startup servo mode to either center at startup or to use custom positions stored in EEPROM.

        Parameters:
            servo_mode:     The new servo mode, either as 0 for center or 1 for custom from EEPROM.

        Raises:
            ValueError if input servo_mode is not valid.
            AssertionError if returned servo_mode_set is not the same as the set servo mode.
        """
        self._check_is_open()
        if servo_mode not in [0, 1]:
            raise ValueError(f"Invalid servo mode {servo_mode}. The valid values are 0 for custom or 1 for EEPROM.")

        servo_mode_set = self._ask("EDD", servo_mode)
        assert servo_mode_set == servo_mode

    @rpc_method
    def clear(self) -> None:
        """This command will set all custom settings, like Port Mode, Servo Disabled, Startup Mode and Default
        Positions, of the servo back to default.

        Raises:
            AssertionError if returned message after clear is not "CLR".
        """
        self._check_is_open()
        response = self._ask("LEAR")
        assert response == "CLR"

    @rpc_method
    def move_up(self, channel: int) -> None:
        """Moves the servo to maximal value.

        Parameters:
            channel:        The channel number.
        """
        self._check_is_open()
        self._check_channel(channel)
        self.set_target_value(channel, self.maximums[channel])

    @rpc_method
    def move_down(self, channel: int) -> None:
        """Moves the servo to minimal value.

        Parameters:
            channel:        The channel number.
        """
        self._check_is_open()
        self._check_channel(channel)
        self.set_target_value(channel, self.minimums[channel])

    @rpc_method
    def is_moving(self, channel: int) -> bool:
        """This method checks if the servo is still moving.

        Return:
            True if servo is still moving, otherwise False.
        """
        self._check_is_open()
        self._check_channel(channel)
        value_now = self._ask("RSP", channel)
        if not value_now == self.targets[channel]:
            return True

        return False

    @rpc_method
    def set_min(self, channel: int, value: int) -> None:
        """Replace the minimum value for a channel.

        Parameters:
            channel:        The channel number.
            value:          The new minimum target value.
        """
        self._check_channel(channel)
        if value < 0 or value >= self.maximums[channel]:
            raise ValueError(f"Minimum value {value} invalid, should be in range [0-{self.maximums[channel] - 1}")

        self.minimums[channel] = value

    @rpc_method
    def set_max(self, channel: int, value: int) -> None:
        """Replace the maximum value for a channel.

        Parameters:
            channel:        The channel number.
            value:          The new maximum target value.
        """
        self._check_channel(channel)
        if value <= self.minimums[channel] or value > self.MAX_PULSE_WIDTH:
            raise ValueError(f"Maximum value {value} invalid, should be in range [{self.minimums[channel] + 1}" +
                             f"-{self.MAX_PULSE_WIDTH}]")

        self.maximums[channel] = value



#! /usr/bin/env python3

"""Test basic start/stop functionality of QMI framework.
"""

import unittest

import qmi
import qmi.core.context
import qmi.core.exceptions


class TestQmiBasic(unittest.TestCase):

    def test_start_stop(self):

        # Check qmi.context() throws QMI_NoActiveContextException if non-existing.
        with self.assertRaises(qmi.core.exceptions.QMI_NoActiveContextException):
            ctx = qmi.context()

        # Check qmi.info() works when there is no context.
        s = qmi.info()
        self.assertIsInstance(s, str)

        # Check qmi.stop() prohibited.
        with self.assertRaises(qmi.core.exceptions.QMI_NoActiveContextException):
            qmi.stop()

        # Start QMI.
        qmi.start("my_test_context")

        # Check qmi.context() returns a context.
        ctx = qmi.context()
        self.assertIsInstance(ctx, qmi.core.context.QMI_Context)
        self.assertEqual(ctx.name, "my_test_context")

        # Check qmi.info() works.
        s = qmi.info()
        self.assertIsInstance(s, str)

        # Check qmi.start() now prohibited.
        with self.assertRaises(qmi.core.exceptions.QMI_UsageException):
            qmi.start("my_second_context")

        # stop QMI
        qmi.stop()

        # Check qmi.context() returns None when there is no context.
        with self.assertRaises(qmi.core.exceptions.QMI_NoActiveContextException):
            ctx = qmi.context()

        # Check qmi.info() works when there is no context.
        s = qmi.info()
        self.assertIsInstance(s, str)

        # Check qmi.stop() now prohibited.
        with self.assertRaises(qmi.core.exceptions.QMI_NoActiveContextException):
            qmi.stop()
        

if __name__ == "__main__":
    unittest.main()
